"""Stochastic synthesis UGens (Gendy family)."""

from typing import Any

from ..enums import CalculationRate
from ..synthdef import (
    Default,
    UGen,
    UGenRecursiveInput,
    param,
    ugen,
)


@ugen(ar=True, kr=True)
class Gendy1(UGen):
    amplitude_distribution = param(1)
    duration_distribution = param(1)
    amplitude_scale = param(0.5)
    duration_scale = param(0.5)
    frequency = param(440.0)
    amplitude_parameter_one = param(0.5)
    amplitude_parameter_two = param(0.5)
    duration_parameter_one = param(0.5)
    duration_parameter_two = param(0.5)
    init_cps = param(12)
    knum = param(Default())

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs: UGenRecursiveInput | None,
    ) -> tuple[CalculationRate, dict[str, Any]]:
        if isinstance(kwargs.get("knum"), Default):
            kwargs["knum"] = kwargs.get("init_cps")
        return calculation_rate, kwargs


@ugen(ar=True, kr=True)
class Gendy2(UGen):
    amplitude_distribution = param(1)
    duration_distribution = param(1)
    amplitude_scale = param(0.5)
    duration_scale = param(0.5)
    frequency = param(440.0)
    amplitude_parameter_one = param(0.5)
    amplitude_parameter_two = param(0.5)
    duration_parameter_one = param(0.5)
    duration_parameter_two = param(0.5)
    init_cps = param(12)
    knum = param(Default())
    a = param(1.17)
    c = param(0.31)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs: UGenRecursiveInput | None,
    ) -> tuple[CalculationRate, dict[str, Any]]:
        if isinstance(kwargs.get("knum"), Default):
            kwargs["knum"] = kwargs.get("init_cps")
        return calculation_rate, kwargs


@ugen(ar=True, kr=True)
class Gendy3(UGen):
    amplitude_distribution = param(1)
    duration_distribution = param(1)
    amplitude_scale = param(0.5)
    duration_scale = param(0.5)
    frequency = param(440.0)
    amplitude_parameter_one = param(0.5)
    amplitude_parameter_two = param(0.5)
    duration_parameter_one = param(0.5)
    duration_parameter_two = param(0.5)
    init_cps = param(12)
    knum = param(Default())

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs: UGenRecursiveInput | None,
    ) -> tuple[CalculationRate, dict[str, Any]]:
        if isinstance(kwargs.get("knum"), Default):
            kwargs["knum"] = kwargs.get("init_cps")
        return calculation_rate, kwargs



"""High-level Server class wrapping the embedded scsynth engine."""

from __future__ import annotations

import contextlib
import itertools
import logging
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any

from .osc import OscMessage
from .scsynth import BootStatus, EmbeddedProcessProtocol, Options

if TYPE_CHECKING:
    from .osc import OscArgument
    from .synthdef import SynthDef

logger = logging.getLogger(__name__)


class Server:
    """High-level wrapper around the embedded scsynth engine.

    Manages the full boot-send-quit lifecycle, node ID allocation,
    SynthDef dispatch, and common OSC commands.

    Can be used as a context manager::

        with Server() as s:
            s.send_synthdef(my_synthdef)
            node = s.synth("my_synth", frequency=880.0)
            ...
    """

    def __init__(self, options: Options | None = None) -> None:
        self._options = options or Options()
        self._protocol = EmbeddedProcessProtocol()
        self._node_id_counter = itertools.count(1000)
        self._synthdefs: set[str] = set()

    def __enter__(self) -> Server:
        self.boot()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        self.quit()

    def __repr__(self) -> str:
        status = "running" if self.is_running else "stopped"
        return f"<Server ({status})>"

    # -- Lifecycle -------------------------------------------------------------

    def boot(self) -> None:
        """Boot the embedded scsynth engine and create the default group."""
        self._protocol.boot(self._options)
        # Create the default group (group 1, add to head of root node 0)
        self.send_msg("/g_new", 1, 0, 0)

    def quit(self) -> None:
        """Shut down the embedded scsynth engine."""
        if not self.is_running:
            return
        self.send_msg("/quit")
        self._protocol._shutdown()

    @property
    def is_running(self) -> bool:
        """Whether the engine is currently online."""
        return self._protocol.status == BootStatus.ONLINE

    # -- Node ID allocation ----------------------------------------------------

    def next_node_id(self) -> int:
        """Return a unique node ID (monotonically increasing from 1000)."""
        return next(self._node_id_counter)

    # -- OSC -------------------------------------------------------------------

    def send_msg(self, address: str, *args: OscArgument) -> None:
        """Send an OSC message to the engine."""
        self._protocol.send_packet(OscMessage(address, *args).to_datagram())

    # -- SynthDef management ---------------------------------------------------

    def send_synthdef(self, synthdef: SynthDef) -> None:
        """Send a compiled SynthDef to the engine via /d_recv."""
        name = synthdef.effective_name
        compiled = synthdef.compile()
        self.send_msg("/d_recv", compiled)
        self._synthdefs.add(name)

    # -- Convenience -----------------------------------------------------------

    def synth(
        self,
        name: str,
        target: int = 1,
        action: int = 0,
        **params: float,
    ) -> int:
        """Create a synth node. Returns the allocated node ID.

        Args:
            name: SynthDef name.
            target: Target node for placement.
            action: Add action (0=head, 1=tail, 2=before, 3=after, 4=replace).
            **params: Initial synth parameter values.
        """
        node_id = self.next_node_id()
        args: list[OscArgument] = [name, node_id, action, target]
        for key, value in params.items():
            args.append(key)
            args.append(float(value))
        self.send_msg("/s_new", *args)
        return node_id

    def group(self, target: int = 0, action: int = 0) -> int:
        """Create a group node. Returns the allocated node ID.

        Args:
            target: Target node for placement.
            action: Add action (0=head, 1=tail, 2=before, 3=after, 4=replace).
        """
        node_id = self.next_node_id()
        self.send_msg("/g_new", node_id, action, target)
        return node_id

    def free(self, node_id: int) -> None:
        """Free a node by ID."""
        self.send_msg("/n_free", node_id)

    @contextlib.contextmanager
    def managed_synth(
        self,
        name: str,
        target: int = 1,
        action: int = 0,
        **params: float,
    ) -> Iterator[int]:
        """Create a synth and free it on context exit.

        Usage::

            with server.managed_synth("sine", frequency=440.0) as node_id:
                time.sleep(1)
            # node freed automatically
        """
        node_id = self.synth(name, target=target, action=action, **params)
        try:
            yield node_id
        finally:
            if self.is_running:
                self.free(node_id)

    @contextlib.contextmanager
    def managed_group(
        self,
        target: int = 0,
        action: int = 0,
    ) -> Iterator[int]:
        """Create a group and free it on context exit."""
        node_id = self.group(target=target, action=action)
        try:
            yield node_id
        finally:
            if self.is_running:
                self.free(node_id)

    def set(self, node_id: int, **params: float) -> None:
        """Set parameter values on a running node.

        Args:
            node_id: The node to modify.
            **params: Parameter name-value pairs.
        """
        args: list[OscArgument] = [node_id]
        for key, value in params.items():
            args.append(key)
            args.append(float(value))
        self.send_msg("/n_set", *args)



"""
01_sine.py -- Hello sine wave.

Builds a SynthDef with SinOsc -> Pan2 -> Out, boots an embedded scsynth,
sends the SynthDef via /d_recv, creates a synth with /s_new, waits, then
cleans up with /quit.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
  - SC_PLUGIN_PATH set to the SuperCollider plugins directory
"""

import time

from nanosynth import OscMessage, Options
from nanosynth.scsynth import _options_to_world_kwargs
from nanosynth._scsynth import (
    world_new,
    world_open_udp,
    world_send_packet,
    world_wait_for_quit,
)
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import Out, Pan2, SinOsc


def send(world, *args):
    """Send an OSC message to the embedded world."""
    world_send_packet(world, OscMessage(*args).to_datagram())


def main():
    # -- Build SynthDef -------------------------------------------------------
    with SynthDefBuilder(frequency=440.0, amplitude=0.3) as builder:
        sig = SinOsc.ar(frequency=builder["frequency"])
        sig = sig * builder["amplitude"]
        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.1, sustain_time=1.8, release_time=0.1
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    synthdef = builder.build(name="sine")
    synthdef_bytes = synthdef.compile()
    print(f"SynthDef '{synthdef.name}' compiled: {len(synthdef_bytes)} bytes")

    # -- Boot embedded scsynth ------------------------------------------------
    world = world_new(
        **_options_to_world_kwargs(Options(verbosity=0, load_synthdefs=False))
    )
    world_open_udp(world, "127.0.0.1", 57110)
    print("Embedded scsynth booted.")

    # Create default group (group 1 inside root group 0)
    send(world, "/g_new", 1, 0, 0)

    # -- Send SynthDef and create synth ---------------------------------------
    send(world, "/d_recv", synthdef_bytes)
    time.sleep(0.1)

    send(world, "/s_new", "sine", 1000, 0, 1, "frequency", 440.0, "amplitude", 0.3)
    print("Playing 440 Hz sine for 2 seconds...")
    time.sleep(2.5)

    # -- Cleanup --------------------------------------------------------------
    send(world, "/quit")
    world_wait_for_quit(world, False)
    print("Done.")


if __name__ == "__main__":
    main()



"""
05_pluck.py -- Karplus-Strong plucked string.

Uses the Pluck UGen (Karplus-Strong algorithm): a burst of noise is fed
into a tuned delay line with feedback to simulate a plucked string.
Dust generates random trigger impulses, producing an evolving plucked
texture at different pitches.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import OscMessage, Options
from nanosynth.scsynth import _options_to_world_kwargs
from nanosynth._scsynth import (
    world_new,
    world_open_udp,
    world_send_packet,
    world_wait_for_quit,
)
from nanosynth.synthdef import SynthDefBuilder
from nanosynth.ugens import Dust, Out, Pan2, Pluck, WhiteNoise


def send(world, *args):
    world_send_packet(world, OscMessage(*args).to_datagram())


def main():
    # -- SynthDef: plucked string with random triggers ------------------------
    with SynthDefBuilder(
        frequency=440.0, decay=4.0, coef=0.3, density=2.0, amplitude=0.5
    ) as builder:
        trig = Dust.ar(density=builder["density"])
        sig = Pluck.ar(
            source=WhiteNoise.ar(),
            trigger=trig,
            maximum_delay_time=1.0 / 60.0,  # lowest pitch ~60 Hz
            delay_time=1.0 / builder["frequency"],
            decay_time=builder["decay"],
            coefficient=builder["coef"],
        )
        sig = sig * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    synthdef = builder.build(name="pluck")
    synthdef_bytes = synthdef.compile()
    print(f"SynthDef '{synthdef.name}' compiled: {len(synthdef_bytes)} bytes")

    # -- Boot and play --------------------------------------------------------
    world = world_new(
        **_options_to_world_kwargs(Options(verbosity=0, load_synthdefs=False))
    )
    world_open_udp(world, "127.0.0.1", 57110)
    print("Embedded scsynth booted.")

    send(world, "/g_new", 1, 0, 0)
    send(world, "/d_recv", synthdef_bytes)
    time.sleep(0.1)

    # Layer several plucked strings at different pitches
    notes = [
        (196.00, -0.3, 1000),  # G3, panned left
        (261.63, 0.0, 1001),  # C4, center
        (329.63, 0.3, 1002),  # E4, panned right
    ]
    print("Playing layered plucked strings (5s)...")
    for freq, pan, node_id in notes:
        send(
            world,
            "/s_new",
            "pluck",
            node_id,
            0,
            1,
            "frequency",
            freq,
            "decay",
            3.0,
            "coef",
            0.2,
            "density",
            1.5,
            "amplitude",
            0.35,
        )
        time.sleep(0.2)

    time.sleep(5.0)

    # Free nodes and quit
    for _, _, node_id in notes:
        send(world, "/n_free", node_id)
    time.sleep(0.1)

    send(world, "/quit")
    world_wait_for_quit(world, False)
    print("Done.")


if __name__ == "__main__":
    main()



"""Tests for the high-level Server class."""

from unittest.mock import MagicMock, patch

import pytest

from nanosynth.scsynth import BootStatus, Options
from nanosynth.server import Server


class TestServerInit:
    def test_default_options(self) -> None:
        """Server() uses default Options when none provided."""
        s = Server()
        assert s._options == Options()

    def test_custom_options(self) -> None:
        """Server(options=...) stores the provided Options."""
        opts = Options(port=57111, verbosity=1)
        s = Server(options=opts)
        assert s._options is opts

    def test_repr_stopped(self) -> None:
        s = Server()
        assert "stopped" in repr(s)

    def test_is_running_false_initially(self) -> None:
        s = Server()
        assert s.is_running is False


class TestNodeIdAllocation:
    def test_starts_at_1000(self) -> None:
        s = Server()
        assert s.next_node_id() == 1000

    def test_monotonically_increasing(self) -> None:
        s = Server()
        ids = [s.next_node_id() for _ in range(5)]
        assert ids == [1000, 1001, 1002, 1003, 1004]


class TestServerWithMockedProtocol:
    """Tests using a mocked EmbeddedProcessProtocol to avoid booting audio."""

    @pytest.fixture()
    def server(self) -> Server:
        s = Server()
        s._protocol = MagicMock()
        s._protocol.status = BootStatus.ONLINE
        return s

    def test_send_msg(self, server: Server) -> None:
        server.send_msg("/test", 1, 2.0, "hello")
        server._protocol.send_packet.assert_called_once()
        data = server._protocol.send_packet.call_args[0][0]
        assert isinstance(data, bytes)

    def test_is_running_when_online(self, server: Server) -> None:
        assert server.is_running is True

    def test_send_synthdef(self, server: Server) -> None:
        sd = MagicMock()
        sd.effective_name = "test_sd"
        sd.compile.return_value = b"SCgf_data"
        server.send_synthdef(sd)
        assert "test_sd" in server._synthdefs
        server._protocol.send_packet.assert_called_once()

    def test_synth_returns_node_id(self, server: Server) -> None:
        node_id = server.synth("my_synth", frequency=440.0)
        assert node_id == 1000
        server._protocol.send_packet.assert_called_once()

    def test_synth_increments_id(self, server: Server) -> None:
        id1 = server.synth("s1")
        id2 = server.synth("s2")
        assert id2 == id1 + 1

    def test_group_returns_node_id(self, server: Server) -> None:
        node_id = server.group(target=1, action=0)
        assert node_id == 1000

    def test_free_sends_n_free(self, server: Server) -> None:
        server.free(1000)
        server._protocol.send_packet.assert_called_once()

    def test_set_sends_n_set(self, server: Server) -> None:
        server.set(1000, frequency=880.0, amplitude=0.5)
        server._protocol.send_packet.assert_called_once()

    def test_quit_when_not_running(self) -> None:
        """quit() is a no-op when not running."""
        s = Server()
        s._protocol = MagicMock()
        s._protocol.status = BootStatus.OFFLINE
        s.quit()
        s._protocol.send_packet.assert_not_called()
        s._protocol._shutdown.assert_not_called()

    def test_context_manager(self) -> None:
        """Context manager calls boot on enter and quit on exit."""
        s = Server()
        s._protocol = MagicMock()
        s._protocol.status = BootStatus.ONLINE
        with patch.object(s, "boot") as mock_boot, patch.object(s, "quit") as mock_quit:
            with s:
                mock_boot.assert_called_once()
            mock_quit.assert_called_once()


class TestServerOscEncoding:
    """Verify that OSC messages are correctly encoded."""

    @pytest.fixture()
    def server(self) -> Server:
        s = Server()
        s._protocol = MagicMock()
        s._protocol.status = BootStatus.ONLINE
        return s

    def test_synth_osc_message(self, server: Server) -> None:
        """synth() sends /s_new with correct argument order."""
        from nanosynth.osc import OscMessage

        server.synth("sine", target=1, action=0, frequency=440.0)
        data = server._protocol.send_packet.call_args[0][0]
        msg = OscMessage.from_datagram(data)
        assert msg.address == "/s_new"
        assert msg.contents[0] == "sine"
        assert msg.contents[1] == 1000  # node id
        assert msg.contents[2] == 0  # action
        assert msg.contents[3] == 1  # target

    def test_group_osc_message(self, server: Server) -> None:
        """group() sends /g_new with correct arguments."""
        from nanosynth.osc import OscMessage

        server.group(target=1, action=1)
        data = server._protocol.send_packet.call_args[0][0]
        msg = OscMessage.from_datagram(data)
        assert msg.address == "/g_new"
        assert msg.contents[0] == 1000
        assert msg.contents[1] == 1
        assert msg.contents[2] == 1

    def test_set_osc_message(self, server: Server) -> None:
        """set() sends /n_set with flattened key-value pairs."""
        from nanosynth.osc import OscMessage

        server.set(1000, frequency=880.0)
        data = server._protocol.send_packet.call_args[0][0]
        msg = OscMessage.from_datagram(data)
        assert msg.address == "/n_set"
        assert msg.contents[0] == 1000
        assert msg.contents[1] == "frequency"
        # OSC float encoding
        assert abs(msg.contents[2] - 880.0) < 0.01


class TestManagedSynth:
    """Tests for managed_synth and managed_group context managers."""

    @pytest.fixture()
    def server(self) -> Server:
        s = Server()
        s._protocol = MagicMock()
        s._protocol.status = BootStatus.ONLINE
        return s

    def test_managed_synth_yields_node_id(self, server: Server) -> None:
        with server.managed_synth("test") as node_id:
            assert node_id == 1000

    def test_managed_synth_frees_on_exit(self, server: Server) -> None:
        from nanosynth.osc import OscMessage

        with server.managed_synth("test") as node_id:
            pass
        # Last call should be /n_free
        last_data = server._protocol.send_packet.call_args[0][0]
        msg = OscMessage.from_datagram(last_data)
        assert msg.address == "/n_free"
        assert msg.contents[0] == node_id

    def test_managed_synth_frees_on_exception(self, server: Server) -> None:
        from nanosynth.osc import OscMessage

        with pytest.raises(RuntimeError):
            with server.managed_synth("test") as node_id:
                raise RuntimeError("boom")
        last_data = server._protocol.send_packet.call_args[0][0]
        msg = OscMessage.from_datagram(last_data)
        assert msg.address == "/n_free"
        assert msg.contents[0] == node_id

    def test_managed_synth_skips_free_if_server_stopped(self) -> None:
        """If the server has already quit, don't try to free."""
        s = Server()
        s._protocol = MagicMock()
        s._protocol.status = BootStatus.ONLINE
        with s.managed_synth("test"):
            s._protocol.status = BootStatus.OFFLINE
        # send_packet called once for /s_new, but NOT for /n_free
        assert s._protocol.send_packet.call_count == 1

    def test_managed_synth_forwards_params(self, server: Server) -> None:
        from nanosynth.osc import OscMessage

        with server.managed_synth("sine", frequency=440.0):
            # Check the /s_new message had the param
            first_data = server._protocol.send_packet.call_args_list[0][0][0]
            msg = OscMessage.from_datagram(first_data)
            assert msg.address == "/s_new"
            assert "frequency" in msg.contents

    def test_managed_group_yields_node_id(self, server: Server) -> None:
        with server.managed_group(target=1) as node_id:
            assert node_id == 1000

    def test_managed_group_frees_on_exit(self, server: Server) -> None:
        from nanosynth.osc import OscMessage

        with server.managed_group() as node_id:
            pass
        last_data = server._protocol.send_packet.call_args[0][0]
        msg = OscMessage.from_datagram(last_data)
        assert msg.address == "/n_free"
        assert msg.contents[0] == node_id



"""
07_delay_reverb.py -- Delay lines and reverb.

Demonstrates effect processing with three SynthDefs:
  1. A short percussive source (filtered saw impulse)
  2. Ping-pong style comb delay feedback
  3. FreeVerb for lush tail

The source feeds into the comb delay, which feeds into reverb,
showing how to chain effects via audio buses.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import OscMessage, Options
from nanosynth.scsynth import _options_to_world_kwargs
from nanosynth._scsynth import (
    world_new,
    world_open_udp,
    world_send_packet,
    world_wait_for_quit,
)
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import (
    AllpassC,
    CombC,
    FreeVerb,
    In,
    Out,
    Pan2,
    RLPF,
    Saw,
)


def send(world, *args):
    world_send_packet(world, OscMessage(*args).to_datagram())


# First private audio bus (after hardware I/O buses)
EFFECT_BUS = 16


def main():
    # -- SynthDef 1: percussive source -> effect bus --------------------------
    with SynthDefBuilder(frequency=440.0, amplitude=0.4) as builder:
        sig = Saw.ar(frequency=builder["frequency"])
        sig = RLPF.ar(source=sig, frequency=2000.0, reciprocal_of_q=0.3)
        env = EnvGen.kr(
            envelope=Envelope.percussive(attack_time=0.003, release_time=0.15),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        # Write to both hardware out (dry) and effect bus
        Out.ar(bus=0, source=Pan2.ar(source=sig * 0.4))
        Out.ar(bus=EFFECT_BUS, source=sig)

    src_def = builder.build(name="perc_src")
    src_bytes = src_def.compile()
    print(f"SynthDef '{src_def.name}' compiled: {len(src_bytes)} bytes")

    # -- SynthDef 2: comb delay effect (reads effect bus) ---------------------
    with SynthDefBuilder(delay_time=0.375, decay_time=3.0, mix=0.5) as builder:
        dry = In.ar(bus=EFFECT_BUS)
        # Two comb delays at slightly different times for stereo width
        left = CombC.ar(
            source=dry,
            maximum_delay_time=1.0,
            delay_time=builder["delay_time"],
            decay_time=builder["decay_time"],
        )
        right = CombC.ar(
            source=dry,
            maximum_delay_time=1.0,
            delay_time=builder["delay_time"] * 0.75,
            decay_time=builder["decay_time"],
        )
        # Allpass diffusion
        left = AllpassC.ar(
            source=left,
            maximum_delay_time=0.05,
            delay_time=0.031,
            decay_time=1.0,
        )
        right = AllpassC.ar(
            source=right,
            maximum_delay_time=0.05,
            delay_time=0.043,
            decay_time=1.0,
        )
        wet = builder["mix"]
        Out.ar(bus=0, source=[left * wet, right * wet])

    delay_def = builder.build(name="comb_delay")
    delay_bytes = delay_def.compile()
    print(f"SynthDef '{delay_def.name}' compiled: {len(delay_bytes)} bytes")

    # -- SynthDef 3: reverb tail (reads hardware out, adds reverb) ------------
    with SynthDefBuilder(room=0.85, damp=0.4, mix=0.3) as builder:
        sig = In.ar(bus=0, channel_count=2)
        left = sig[0]
        right = sig[1]
        left = FreeVerb.ar(
            source=left,
            mix=builder["mix"],
            room_size=builder["room"],
            damping=builder["damp"],
        )
        right = FreeVerb.ar(
            source=right,
            mix=builder["mix"],
            room_size=builder["room"],
            damping=builder["damp"],
        )
        Out.ar(bus=0, source=[left, right])

    verb_def = builder.build(name="reverb")
    verb_bytes = verb_def.compile()
    print(f"SynthDef '{verb_def.name}' compiled: {len(verb_bytes)} bytes")

    # -- Boot and play --------------------------------------------------------
    world = world_new(
        **_options_to_world_kwargs(Options(verbosity=0, load_synthdefs=False))
    )
    world_open_udp(world, "127.0.0.1", 57110)
    print("Embedded scsynth booted.")

    send(world, "/g_new", 1, 0, 0)  # source group
    send(world, "/g_new", 2, 3, 1)  # effect group (after source group)

    send(world, "/d_recv", src_bytes)
    send(world, "/d_recv", delay_bytes)
    send(world, "/d_recv", verb_bytes)
    time.sleep(0.1)

    # Start persistent effect synths in the effect group
    send(
        world,
        "/s_new",
        "comb_delay",
        100,
        0,
        2,
        "delay_time",
        0.375,
        "decay_time",
        4.0,
        "mix",
        0.4,
    )
    send(world, "/s_new", "reverb", 101, 3, 2, "room", 0.8, "damp", 0.5, "mix", 0.25)

    # Fire percussive notes in the source group
    melody = [
        (329.63, 0.375),  # E4
        (392.00, 0.375),  # G4
        (440.00, 0.375),  # A4
        (523.25, 0.375),  # C5
        (440.00, 0.375),  # A4
        (392.00, 0.375),  # G4
        (329.63, 0.750),  # E4 (held longer)
        (293.66, 0.750),  # D4
    ]

    print("Playing percussive melody through delay + reverb...")
    node_id = 1000
    for freq, dur in melody:
        send(
            world,
            "/s_new",
            "perc_src",
            node_id,
            0,
            1,
            "frequency",
            freq,
            "amplitude",
            0.5,
        )
        time.sleep(dur)
        node_id += 1

    # Let the delay tail ring out
    print("Letting delay tail ring out (4s)...")
    time.sleep(4.0)

    send(world, "/quit")
    world_wait_for_quit(world, False)
    print("Done.")


if __name__ == "__main__":
    main()



"""
11_stereo_pad.py -- Lush detuned stereo pad with MoogFF filter.

Builds a thick stereo pad by layering detuned Pulse oscillators with
varying widths, run through a MoogFF (Moog-style ladder filter) with
an LFO-modulated cutoff. Demonstrates:
  - Detuning for stereo width
  - Pulse width as a timbral control
  - MoogFF resonant filter
  - Slow LFO modulation at control rate
  - Multi-voice layering within a single SynthDef

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import OscMessage, Options
from nanosynth.scsynth import _options_to_world_kwargs
from nanosynth._scsynth import (
    world_new,
    world_open_udp,
    world_send_packet,
    world_wait_for_quit,
)
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import LFNoise2, MoogFF, Out, Pulse


def send(world, *args):
    world_send_packet(world, OscMessage(*args).to_datagram())


def main():
    # -- SynthDef: detuned stereo pad with MoogFF -----------------------------
    with SynthDefBuilder(
        frequency=220.0,
        detune=0.8,
        cutoff=1200.0,
        resonance=2.5,
        amplitude=0.3,
        gate=1.0,
    ) as builder:
        freq = builder["frequency"]
        det = builder["detune"]

        # Left channel: two detuned pulse oscillators
        left = Pulse.ar(frequency=freq - det, width=0.3) * 0.5
        left = left + Pulse.ar(frequency=freq - det * 2.0, width=0.5) * 0.5

        # Right channel: two detuned pulse oscillators (opposite detune)
        right = Pulse.ar(frequency=freq + det, width=0.3) * 0.5
        right = right + Pulse.ar(frequency=freq + det * 2.0, width=0.5) * 0.5

        # Slowly wandering cutoff modulation
        lfo = LFNoise2.kr(frequency=0.2)
        cutoff = builder["cutoff"] + lfo * 600.0

        # MoogFF ladder filter on each channel
        left = MoogFF.ar(source=left, frequency=cutoff, gain=builder["resonance"])
        right = MoogFF.ar(source=right, frequency=cutoff, gain=builder["resonance"])

        # ADSR envelope with gate control
        env = EnvGen.kr(
            envelope=Envelope.adsr(
                attack_time=1.5,
                decay_time=0.5,
                sustain=0.8,
                release_time=2.0,
            ),
            gate=builder["gate"],
            done_action=DoneAction.FREE_SYNTH,
        )
        left = left * env * builder["amplitude"]
        right = right * env * builder["amplitude"]
        Out.ar(bus=0, source=[left, right])

    synthdef = builder.build(name="stereo_pad")
    synthdef_bytes = synthdef.compile()
    print(f"SynthDef '{synthdef.name}' compiled: {len(synthdef_bytes)} bytes")

    # -- Boot and play --------------------------------------------------------
    world = world_new(
        **_options_to_world_kwargs(Options(verbosity=0, load_synthdefs=False))
    )
    world_open_udp(world, "127.0.0.1", 57110)
    print("Embedded scsynth booted.")

    send(world, "/g_new", 1, 0, 0)
    send(world, "/d_recv", synthdef_bytes)
    time.sleep(0.1)

    # Play a chord progression: Am -> F -> C -> G
    chords = [
        (220.00, "Am"),  # A3
        (174.61, "F"),  # F3
        (261.63, "C"),  # C4
        (196.00, "G"),  # G3
    ]

    print("Playing pad chord progression...")
    node_id = 1000
    for freq, name in chords:
        print(f"  {name} ({freq:.0f} Hz)...")
        send(
            world,
            "/s_new",
            "stereo_pad",
            node_id,
            0,
            1,
            "frequency",
            freq,
            "detune",
            0.7,
            "cutoff",
            1000.0,
            "resonance",
            2.0,
            "amplitude",
            0.25,
        )
        time.sleep(3.0)
        # Release the gate
        send(world, "/n_set", node_id, "gate", 0.0)
        time.sleep(1.0)
        node_id += 1

    # Let the last release tail finish
    time.sleep(2.0)

    send(world, "/quit")
    world_wait_for_quit(world, False)
    print("Done.")


if __name__ == "__main__":
    main()



"""
04_additive.py -- Additive synthesis.

Builds a tone by summing sine partials with decreasing amplitudes.
An LFO slowly modulates the balance between odd and even harmonics,
creating a timbral shift from hollow (odd-only, like a square wave)
to bright (all harmonics).

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import OscMessage, Options
from nanosynth.scsynth import _options_to_world_kwargs
from nanosynth._scsynth import (
    world_new,
    world_open_udp,
    world_send_packet,
    world_wait_for_quit,
)
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import LFTri, Out, Pan2, SinOsc


def send(world, *args):
    world_send_packet(world, OscMessage(*args).to_datagram())


def main():
    # -- SynthDef: additive with LFO-morphing harmonics -----------------------
    with SynthDefBuilder(frequency=200.0, amplitude=0.3) as builder:
        freq = builder["frequency"]

        # LFO: 0..1 controls even-harmonic level (odd harmonics always present)
        even_mix = LFTri.kr(frequency=0.15) * 0.5 + 0.5

        # Odd harmonics (1, 3, 5, 7)
        sig = SinOsc.ar(frequency=freq) * 1.0
        sig = sig + SinOsc.ar(frequency=freq * 3.0) * 0.33
        sig = sig + SinOsc.ar(frequency=freq * 5.0) * 0.2
        sig = sig + SinOsc.ar(frequency=freq * 7.0) * 0.14

        # Even harmonics (2, 4, 6) scaled by LFO
        sig = sig + SinOsc.ar(frequency=freq * 2.0) * 0.5 * even_mix
        sig = sig + SinOsc.ar(frequency=freq * 4.0) * 0.25 * even_mix
        sig = sig + SinOsc.ar(frequency=freq * 6.0) * 0.17 * even_mix

        sig = sig * 0.2  # normalize headroom
        env = EnvGen.kr(
            envelope=Envelope.linen(
                attack_time=0.5, sustain_time=5.0, release_time=1.0
            ),
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = sig * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    synthdef = builder.build(name="additive")
    synthdef_bytes = synthdef.compile()
    print(f"SynthDef '{synthdef.name}' compiled: {len(synthdef_bytes)} bytes")

    # -- Boot and play --------------------------------------------------------
    world = world_new(
        **_options_to_world_kwargs(Options(verbosity=0, load_synthdefs=False))
    )
    world_open_udp(world, "127.0.0.1", 57110)
    print("Embedded scsynth booted.")

    send(world, "/g_new", 1, 0, 0)
    send(world, "/d_recv", synthdef_bytes)
    time.sleep(0.1)

    print("Playing additive synthesis with morphing harmonics (6.5s)...")
    send(world, "/s_new", "additive", 1000, 0, 1, "frequency", 150.0, "amplitude", 0.4)
    time.sleep(7.0)

    send(world, "/quit")
    world_wait_for_quit(world, False)
    print("Done.")


if __name__ == "__main__":
    main()



"""nanosynth -- minimal embedded SuperCollider synthesis engine wrapper."""

__version__ = "0.1.2"

from .enums import CalculationRate, DoneAction
from .osc import OscBundle, OscMessage, find_free_port
from .scsynth import EmbeddedProcessProtocol, Options, find_ugen_plugins_path
from .compiler import compile_synthdefs
from .synthdef import (
    Default,
    Param,
    PseudoUGen,
    SynthDef,
    SynthDefBuilder,
    UGen,
    param,
    synthdef,
    ugen,
)
from .envelopes import EnvGen, Envelope
from .server import Server
from .ugens import *  # noqa: F403
from .ugens import __all__ as _ugen_names

__all__ = [
    "CalculationRate",
    "Default",
    "DoneAction",
    "EmbeddedProcessProtocol",
    "EnvGen",
    "Envelope",
    "OscBundle",
    "OscMessage",
    "Options",
    "Param",
    "PseudoUGen",
    "Server",
    "SynthDef",
    "SynthDefBuilder",
    "UGen",
    "compile_synthdefs",
    "find_free_port",
    "find_ugen_plugins_path",
    "param",
    "synthdef",
    "ugen",
] + _ugen_names



"""
03_fm.py -- FM synthesis + melody.

Two SynthDefs:
  1. FM synth with controllable carrier_freq, mod_ratio, mod_index, gate
     (uses Envelope.adsr for amplitude shaping)
  2. FM sweep with XLine-modulated index (self-freeing)

Plays a short melody with the gated FM synth, then fires a self-freeing sweep.

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
  - SC_PLUGIN_PATH set to the SuperCollider plugins directory
"""

import time

from nanosynth import OscMessage, Options
from nanosynth.scsynth import _options_to_world_kwargs
from nanosynth._scsynth import (
    world_new,
    world_open_udp,
    world_send_packet,
    world_wait_for_quit,
)
from nanosynth.envelopes import EnvGen, Envelope
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import Out, Pan2, SinOsc, XLine


def send(world, *args):
    world_send_packet(world, OscMessage(*args).to_datagram())


def main():
    # -- SynthDef 1: gated FM synth -------------------------------------------
    with SynthDefBuilder(
        carrier_freq=440.0,
        mod_ratio=2.0,
        mod_index=3.0,
        amplitude=0.3,
        gate=1.0,
    ) as builder:
        mod_freq = builder["carrier_freq"] * builder["mod_ratio"]
        modulator = SinOsc.ar(frequency=mod_freq) * builder["mod_index"] * mod_freq
        carrier = SinOsc.ar(frequency=builder["carrier_freq"] + modulator)
        env = EnvGen.kr(
            envelope=Envelope.adsr(
                attack_time=0.01,
                decay_time=0.1,
                sustain=0.7,
                release_time=0.3,
            ),
            gate=builder["gate"],
            done_action=DoneAction.FREE_SYNTH,
        )
        sig = carrier * env * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    fm_def = builder.build(name="fm_synth")
    fm_bytes = fm_def.compile()
    print(f"SynthDef '{fm_def.name}' compiled: {len(fm_bytes)} bytes")

    # -- SynthDef 2: self-freeing FM sweep ------------------------------------
    with SynthDefBuilder(carrier_freq=200.0, mod_ratio=3.0, amplitude=0.25) as builder:
        mod_index = XLine.kr(
            start=10.0,
            stop=0.1,
            duration=4.0,
            done_action=DoneAction.FREE_SYNTH,
        )
        mod_freq = builder["carrier_freq"] * builder["mod_ratio"]
        modulator = SinOsc.ar(frequency=mod_freq) * mod_index * mod_freq
        carrier = SinOsc.ar(frequency=builder["carrier_freq"] + modulator)
        sig = carrier * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig))

    sweep_def = builder.build(name="fm_sweep")
    sweep_bytes = sweep_def.compile()
    print(f"SynthDef '{sweep_def.name}' compiled: {len(sweep_bytes)} bytes")

    # -- Boot and play --------------------------------------------------------
    world = world_new(
        **_options_to_world_kwargs(Options(verbosity=0, load_synthdefs=False))
    )
    world_open_udp(world, "127.0.0.1", 57110)
    print("Embedded scsynth booted.")

    # Create default group (group 1 inside root group 0)
    send(world, "/g_new", 1, 0, 0)

    send(world, "/d_recv", fm_bytes)
    send(world, "/d_recv", sweep_bytes)
    time.sleep(0.1)

    # Play a short melody with the gated FM synth
    melody = [
        (261.63, 0.3),  # C4
        (293.66, 0.3),  # D4
        (329.63, 0.3),  # E4
        (349.23, 0.3),  # F4
        (392.00, 0.5),  # G4
        (440.00, 0.5),  # A4
        (493.88, 0.5),  # B4
        (523.25, 0.8),  # C5
    ]

    print("Playing FM melody...")
    node_id = 1000
    for freq, dur in melody:
        send(
            world,
            "/s_new",
            "fm_synth",
            node_id,
            0,
            1,
            "carrier_freq",
            freq,
            "mod_ratio",
            2.0,
            "mod_index",
            3.0,
            "amplitude",
            0.25,
        )
        time.sleep(dur)
        # Release the gate to trigger the envelope release
        send(world, "/n_set", node_id, "gate", 0.0)
        time.sleep(0.05)
        node_id += 1

    time.sleep(0.5)

    # Fire the self-freeing FM sweep
    print("Playing FM sweep (4s)...")
    send(world, "/s_new", "fm_sweep", 2000, 0, 1, "carrier_freq", 150.0)
    time.sleep(4.5)

    send(world, "/quit")
    world_wait_for_quit(world, False)
    print("Done.")


if __name__ == "__main__":
    main()



"""
06_ringmod.py -- Ring modulation and amplitude modulation.

Two SynthDefs demonstrating different modulation techniques:
  1. Ring modulation: carrier * modulator (bipolar, produces sum and
     difference frequencies)
  2. Amplitude modulation: carrier * (modulator * 0.5 + 0.5) (unipolar,
     retains carrier fundamental with sidebands)

Requires:
  - nanosynth built with embedded libscsynth (NANOSYNTH_EMBED_SCSYNTH=ON)
"""

import time

from nanosynth import OscMessage, Options
from nanosynth.scsynth import _options_to_world_kwargs
from nanosynth._scsynth import (
    world_new,
    world_open_udp,
    world_send_packet,
    world_wait_for_quit,
)
from nanosynth.synthdef import DoneAction, SynthDefBuilder
from nanosynth.ugens import LFTri, Out, Pan2, SinOsc, XLine


def send(world, *args):
    world_send_packet(world, OscMessage(*args).to_datagram())


def main():
    # -- SynthDef 1: ring modulation with sweeping mod frequency --------------
    with SynthDefBuilder(carrier_freq=440.0, amplitude=0.3) as builder:
        carrier = SinOsc.ar(frequency=builder["carrier_freq"])
        # Sweep modulator frequency from 2 Hz (tremolo) up to 300 Hz (sidebands)
        mod_freq = XLine.kr(
            start=2.0,
            stop=300.0,
            duration=4.0,
            done_action=DoneAction.FREE_SYNTH,
        )
        modulator = LFTri.ar(frequency=mod_freq)
        sig = carrier * modulator * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig, position=-0.4))

    ring_def = builder.build(name="ring_mod")
    ring_bytes = ring_def.compile()
    print(f"SynthDef '{ring_def.name}' compiled: {len(ring_bytes)} bytes")

    # -- SynthDef 2: amplitude modulation with sweeping mod frequency ---------
    with SynthDefBuilder(carrier_freq=440.0, amplitude=0.3) as builder:
        carrier = SinOsc.ar(frequency=builder["carrier_freq"])
        mod_freq = XLine.kr(
            start=2.0,
            stop=300.0,
            duration=4.0,
            done_action=DoneAction.FREE_SYNTH,
        )
        # Unipolar modulator: 0..1 instead of -1..1
        modulator = SinOsc.ar(frequency=mod_freq) * 0.5 + 0.5
        sig = carrier * modulator * builder["amplitude"]
        Out.ar(bus=0, source=Pan2.ar(source=sig, position=0.4))

    am_def = builder.build(name="am_mod")
    am_bytes = am_def.compile()
    print(f"SynthDef '{am_def.name}' compiled: {len(am_bytes)} bytes")

    # -- Boot and play --------------------------------------------------------
    world = world_new(
        **_options_to_world_kwargs(Options(verbosity=0, load_synthdefs=False))
    )
    world_open_udp(world, "127.0.0.1", 57110)
    print("Embedded scsynth booted.")

    send(world, "/g_new", 1, 0, 0)
    send(world, "/d_recv", ring_bytes)
    send(world, "/d_recv", am_bytes)
    time.sleep(0.1)

    # Ring mod on left, AM on right -- hear the difference
    print("Ring modulation (left) -- sweeping mod freq 2->300 Hz (4s)...")
    send(
        world,
        "/s_new",
        "ring_mod",
        1000,
        0,
        1,
        "carrier_freq",
        440.0,
        "amplitude",
        0.35,
    )
    time.sleep(0.5)

    print("Amplitude modulation (right) -- same sweep...")
    send(
        world, "/s_new", "am_mod", 1001, 0, 1, "carrier_freq", 440.0, "amplitude", 0.35
    )
    time.sleep(4.5)

    send(world, "/quit")
    world_wait_for_quit(world, False)
    print("Done.")


if __name__ == "__main__":
    main()



"""
SynthDef builder and SCgf compiler for nanosynth.

Ported from supriya's ugens/core.py -- minimal subset for defining UGen graphs
in Python and compiling them to SuperCollider's SCgf binary format.
"""

import copy
import enum
import hashlib
import operator
import threading
import uuid
from collections.abc import Sequence as SequenceABC
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    NamedTuple,
    SupportsFloat,
    SupportsInt,
    Union,
    cast,
    overload,
)

from .enums import (  # noqa: F401
    BinaryOperator,
    CalculationRate,
    DoneAction,
    EnvelopeShape,
    ParameterRate,
    UnaryOperator,
)


# ---------------------------------------------------------------------------
# Type aliases
# ---------------------------------------------------------------------------

UGenScalarInput = Union[SupportsFloat, "UGenScalar"]
UGenVectorInput = Union["UGenSerializable", SequenceABC[UGenScalarInput]]
UGenRecursiveInput = Union[
    SupportsFloat,
    "UGenOperable",
    "UGenSerializable",
    SequenceABC["UGenRecursiveInput"],
]
UGenParams = dict[str, Union[UGenScalarInput, UGenVectorInput]]
UGenRecursiveParams = Union[UGenParams, list["UGenRecursiveParams"]]


# ---------------------------------------------------------------------------
# @ugen / @param decorator support
# ---------------------------------------------------------------------------


class Missing:
    """Sentinel for required parameters (no default)."""

    pass


MISSING = Missing()


class Default:
    """Sentinel for parameters whose default is computed from other parameters."""

    def __repr__(self) -> str:
        return "Default()"


class Param(NamedTuple):
    default: "Missing | Default | float | None" = MISSING
    unexpanded: bool = False


def _format_value(value: object) -> str:
    if value == float("inf"):
        value_repr = 'float("inf")'
    elif value == float("-inf"):
        value_repr = 'float("-inf")'
    elif isinstance(value, Missing):
        value_repr = "Missing()"
    elif isinstance(value, Default):
        value_repr = "Default()"
    elif isinstance(value, enum.Enum):
        value_repr = f"{type(value).__name__}.{value.name}"
    else:
        value_repr = repr(value)
    return value_repr


def _get_fn_globals() -> dict[str, Any]:
    return {
        "CalculationRate": CalculationRate,
        "Default": Default,
        "Missing": Missing,
        "SupportsFloat": SupportsFloat,
        "UGenRecursiveInput": UGenRecursiveInput,
        "UGenScalar": UGenScalar,
        "UGenSerializable": UGenSerializable,
        "UGenVector": UGenVector,
        "UGenScalarInput": UGenScalarInput,
        "UGenVectorInput": UGenVectorInput,
        "Union": Union,
    }


def _create_fn(
    *,
    cls: type["UGen"],
    name: str,
    args: list[str],
    body: list[str],
    return_type: Any,
    globals_: dict[str, Any] | None = None,
    decorator: Callable[..., Any] | None = None,
    override: bool = False,
) -> None:
    if name in cls.__dict__ and not override:
        return
    globals_ = globals_ or {}
    locals_ = {"_return_type": return_type}
    args_ = ",\n        ".join(args)
    body_ = "\n".join(f"        {line}" for line in body)
    text = f"    def {name}(\n        {args_}\n    ) -> _return_type:\n{body_}"
    local_vars = ", ".join(locals_.keys())
    text = f"def __create_fn__({local_vars}):\n{text}\n    return {name}"
    namespace: dict[str, Callable[..., Any]] = {}
    exec(text, globals_, namespace)
    value = namespace["__create_fn__"](**locals_)
    value.__qualname__ = f"{cls.__qualname__}.{value.__name__}"
    if decorator:
        value = decorator(value)
    setattr(cls, name, value)


def _add_init(
    cls: type["UGen"],
    params: dict[str, Param],
    is_multichannel: bool,
    channel_count: int,
    fixed_channel_count: bool,
) -> None:
    parent_class = UGen
    args = ["self", "*", "calculation_rate: CalculationRate"]
    body = [
        f"{'self._channel_count = channel_count' if is_multichannel and not fixed_channel_count else ''}"
    ]
    if is_multichannel and fixed_channel_count:
        body = [f"self._channel_count = {channel_count}"]
    elif not is_multichannel and channel_count != 1:
        body = [f"self._channel_count = {channel_count}"]
    else:
        body = []
    body.append(f"{parent_class.__name__}.__init__(")
    body.append("    self,")
    body.append("    calculation_rate=calculation_rate,")
    for key, param in params.items():
        value_repr = _format_value(param.default)
        type_ = "UGenVectorInput" if param.unexpanded else "UGenScalarInput"
        prefix = f"{key}: {type_}"
        args.append(
            f"{prefix} = {value_repr}"
            if not isinstance(param.default, Missing)
            else prefix
        )
        body.append(f"    {key}={key},")
    if is_multichannel and not fixed_channel_count:
        args.append(f"channel_count: int = {channel_count or 1}")
        body.insert(0, "self._channel_count = channel_count")
    args.append("**kwargs")
    body.append("    **kwargs,")
    body.append(")")
    _create_fn(
        cls=cls,
        name="__init__",
        args=args,
        body=body,
        globals_={**_get_fn_globals(), parent_class.__name__: parent_class},
        return_type=None,
    )


def _add_param_fn(cls: type["UGen"], name: str, index: int, unexpanded: bool) -> None:
    _create_fn(
        cls=cls,
        name=name,
        args=["self"],
        body=(
            [f"return self._inputs[{index}:]"]
            if unexpanded
            else [f"return self._inputs[{index}]"]
        ),
        decorator=property,
        globals_=_get_fn_globals(),
        override=True,
        return_type=UGenVector if unexpanded else UGenScalar,
    )


def _add_rate_fn(
    cls: type["UGen"],
    rate: CalculationRate | None,
    params: dict[str, "Param"],
    is_multichannel: bool,
    channel_count: int,
    fixed_channel_count: bool,
) -> None:
    args = ["cls"]
    if params:
        args.append("*")
    for key, param in params.items():
        value_repr = _format_value(param.default)
        prefix = f"{key}: UGenRecursiveInput"
        args.append(
            f"{prefix} = {value_repr}"
            if not isinstance(param.default, Missing)
            else prefix
        )
    body = ["return cls._new_expanded("]
    if rate is None:
        body.append("    calculation_rate=None,")
    else:
        body.append(f"    calculation_rate=CalculationRate.{rate.name},")
    if is_multichannel and not fixed_channel_count:
        args.append(f"channel_count: int = {channel_count or 1}")
        body.append("    channel_count=channel_count,")
    body.extend(f"    {name}={name}," for name in params)
    body.append(")")
    _create_fn(
        cls=cls,
        name=rate.token if rate is not None else "new",
        args=args,
        body=body,
        decorator=classmethod,
        globals_=_get_fn_globals(),
        return_type=UGenOperable,
    )


def _process_class(
    cls: type["UGen"],
    *,
    ar: bool = False,
    kr: bool = False,
    ir: bool = False,
    dr: bool = False,
    new: bool = False,
    has_done_flag: bool = False,
    is_multichannel: bool = False,
    is_output: bool = False,
    is_pure: bool = False,
    is_width_first: bool = False,
    channel_count: int = 1,
    fixed_channel_count: bool = False,
) -> type["UGen"]:
    params: dict[str, Param] = {}
    unexpanded_keys = []
    valid_calculation_rates = []
    for name, value in cls.__dict__.items():
        if not isinstance(value, Param):
            continue
        params[name] = value
        if value.unexpanded:
            unexpanded_keys.append(name)
        _add_param_fn(cls, name, len(params) - 1, value.unexpanded)
    _add_init(cls, params, is_multichannel, channel_count, fixed_channel_count)
    for should_add, rate in [
        (ar, CalculationRate.AUDIO),
        (kr, CalculationRate.CONTROL),
        (ir, CalculationRate.SCALAR),
        (dr, CalculationRate.DEMAND),
        (new, None),
    ]:
        if not should_add:
            continue
        _add_rate_fn(
            cls, rate, params, is_multichannel, channel_count, fixed_channel_count
        )
        if rate is not None:
            valid_calculation_rates.append(rate)
    cls._has_done_flag = bool(has_done_flag)
    cls._is_output = bool(is_output)
    cls._is_pure = bool(is_pure)
    cls._is_width_first = bool(is_width_first)
    cls._ordered_keys = tuple(params.keys())
    cls._unexpanded_keys = frozenset(unexpanded_keys)
    cls._valid_calculation_rates = tuple(valid_calculation_rates)  # type: ignore[attr-defined]
    if cls.__doc__ is None:
        rate_tokens = ", ".join(r.token for r in valid_calculation_rates)
        lines = [f"{cls.__name__} -- {rate_tokens}" if rate_tokens else cls.__name__]
        if params:
            lines.append("")
            lines.append("Parameters:")
            for name, p in params.items():
                if isinstance(p.default, Missing):
                    lines.append(f"    {name} (required)")
                else:
                    lines.append(f"    {name} (default: {p.default})")
        cls.__doc__ = "\n".join(lines)
    return cls


def param(
    default: Missing | Default | float | None = MISSING,
    *,
    unexpanded: bool = False,
) -> Param:
    """Define a UGen parameter. Akin to dataclasses.field."""
    return Param(default, unexpanded)


def ugen(
    *,
    ar: bool = False,
    kr: bool = False,
    ir: bool = False,
    dr: bool = False,
    new: bool = False,
    has_done_flag: bool = False,
    is_multichannel: bool = False,
    is_output: bool = False,
    is_pure: bool = False,
    is_width_first: bool = False,
    channel_count: int = 1,
    fixed_channel_count: bool = False,
) -> Callable[[type["UGen"]], type["UGen"]]:
    """Decorate a UGen class. Akin to dataclasses.dataclass."""

    def wrap(cls: type[UGen]) -> type[UGen]:
        return _process_class(
            cls,
            ar=ar,
            kr=kr,
            ir=ir,
            dr=dr,
            new=new,
            has_done_flag=has_done_flag,
            is_multichannel=is_multichannel,
            is_output=is_output,
            is_pure=is_pure,
            is_width_first=is_width_first,
            channel_count=channel_count,
            fixed_channel_count=fixed_channel_count,
        )

    if is_multichannel and fixed_channel_count:
        raise ValueError(
            "is_multichannel and fixed_channel_count are mutually exclusive"
        )
    return wrap


# ---------------------------------------------------------------------------
# Protocols / ABCs
# ---------------------------------------------------------------------------


class UGenSerializable:
    __slots__ = ()

    def serialize(self) -> "UGenVector":
        raise NotImplementedError


class PseudoUGen:
    """Base class for virtual UGens that compose other UGens."""

    def __init__(self) -> None:
        raise NotImplementedError


# ---------------------------------------------------------------------------
# Helper functions
# ---------------------------------------------------------------------------


def _compute_binary_op(
    left: UGenRecursiveInput,
    right: UGenRecursiveInput,
    special_index: BinaryOperator,
    float_operator: Callable[..., Any] | None = None,
) -> "UGenOperable":
    def recurse(all_expanded_params: UGenRecursiveParams) -> "UGenOperable":
        if not isinstance(all_expanded_params, dict) and len(all_expanded_params) == 1:
            all_expanded_params = all_expanded_params[0]
        if isinstance(all_expanded_params, dict):
            if (
                isinstance(left, SupportsFloat)
                and isinstance(right, SupportsFloat)
                and float_operator is not None
            ):
                return ConstantProxy(float_operator(float(left), float(right)))
            return BinaryOpUGen._new_single(
                calculation_rate=max(
                    [
                        CalculationRate.from_expr(left),
                        CalculationRate.from_expr(right),
                    ]
                ),
                special_index=special_index,
                **all_expanded_params,
            )
        return UGenVector(
            *(recurse(expanded_params) for expanded_params in all_expanded_params)
        )

    return recurse(UGen._expand_params({"left": left, "right": right}))


def _compute_unary_op(
    source: UGenRecursiveInput,
    special_index: UnaryOperator,
    float_operator: Callable[..., Any] | None = None,
) -> "UGenOperable":
    def recurse(all_expanded_params: UGenRecursiveParams) -> "UGenOperable":
        if not isinstance(all_expanded_params, dict) and len(all_expanded_params) == 1:
            all_expanded_params = all_expanded_params[0]
        if isinstance(all_expanded_params, dict):
            if isinstance(source, SupportsFloat) and float_operator is not None:
                return ConstantProxy(float_operator(float(source)))
            return UnaryOpUGen._new_single(
                calculation_rate=max([CalculationRate.from_expr(source)]),
                special_index=special_index,
                **all_expanded_params,
            )
        return UGenVector(
            *(recurse(expanded_params) for expanded_params in all_expanded_params)
        )

    return recurse(UGen._expand_params({"source": source}))


# ---------------------------------------------------------------------------
# Core UGen graph classes
# ---------------------------------------------------------------------------


class UGenOperable:
    """Mixin for UGen arithmetic operations."""

    __slots__ = ()

    def __abs__(self) -> "UGenOperable":
        return _compute_unary_op(
            source=self,
            special_index=UnaryOperator.ABSOLUTE_VALUE,
            float_operator=operator.abs,
        )

    def __add__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=self,
            right=expr,
            special_index=BinaryOperator.ADDITION,
            float_operator=operator.add,
        )

    def __radd__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=expr,
            right=self,
            special_index=BinaryOperator.ADDITION,
            float_operator=operator.add,
        )

    def __sub__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=self,
            right=expr,
            special_index=BinaryOperator.SUBTRACTION,
            float_operator=operator.sub,
        )

    def __rsub__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=expr,
            right=self,
            special_index=BinaryOperator.SUBTRACTION,
            float_operator=operator.sub,
        )

    def __mul__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=self,
            right=expr,
            special_index=BinaryOperator.MULTIPLICATION,
            float_operator=operator.mul,
        )

    def __rmul__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=expr,
            right=self,
            special_index=BinaryOperator.MULTIPLICATION,
            float_operator=operator.mul,
        )

    def __truediv__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=self,
            right=expr,
            special_index=BinaryOperator.FLOAT_DIVISION,
            float_operator=operator.truediv,
        )

    def __rtruediv__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=expr,
            right=self,
            special_index=BinaryOperator.FLOAT_DIVISION,
            float_operator=operator.truediv,
        )

    def __mod__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=self,
            right=expr,
            special_index=BinaryOperator.MODULO,
            float_operator=operator.mod,
        )

    def __gt__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=self,
            right=expr,
            special_index=BinaryOperator.GREATER_THAN,
        )

    def __lt__(self, expr: UGenRecursiveInput) -> "UGenOperable":
        return _compute_binary_op(
            left=self,
            right=expr,
            special_index=BinaryOperator.LESS_THAN,
        )

    def __neg__(self) -> "UGenOperable":
        return _compute_unary_op(
            source=self,
            special_index=UnaryOperator.NEGATIVE,
            float_operator=operator.neg,
        )


class UGenScalar(UGenOperable):
    """A UGen scalar."""

    __slots__ = ()

    def __iter__(self) -> Iterator["UGenOperable"]:
        yield self


class OutputProxy(UGenScalar):
    """A UGen output proxy -- reference to a specific output of a UGen."""

    __slots__ = ("ugen", "index")

    def __init__(self, ugen: "UGen", index: int) -> None:
        self.ugen = ugen
        self.index = index

    def __eq__(self, expr: object) -> bool:
        return (
            isinstance(expr, type(self))
            and self.ugen is expr.ugen
            and self.index == expr.index
        )

    def __hash__(self) -> int:
        return hash((type(self), id(self.ugen), self.index))

    def __repr__(self) -> str:
        return repr(self.ugen).replace(">", f"[{self.index}]>")

    @property
    def calculation_rate(self) -> CalculationRate:
        return self.ugen.calculation_rate


class ConstantProxy(UGenScalar):
    """Wraps a float constant, exposing UGenOperable arithmetic."""

    __slots__ = ("value",)

    def __init__(self, value: SupportsFloat) -> None:
        self.value = float(value)

    def __eq__(self, expr: object) -> bool:
        if isinstance(expr, SupportsFloat):
            return float(self) == float(expr)
        return False

    def __float__(self) -> float:
        return self.value

    def __hash__(self) -> int:
        return hash(self.value)

    def __repr__(self) -> str:
        return f"<{self.value}>"


class UGenVector(UGenOperable, SequenceABC["UGenOperable"]):
    """A sequence of UGenOperables."""

    __slots__ = ("_values",)

    def __init__(self, *values: SupportsFloat | UGenOperable) -> None:
        values_: list[UGen | UGenScalar | "UGenVector"] = []
        for x in values:
            if isinstance(x, (UGen, UGenScalar, UGenVector)):
                values_.append(x)
            elif isinstance(x, UGenSerializable):
                values_.append(UGenVector(*x.serialize()))
            elif isinstance(x, SupportsFloat):
                values_.append(ConstantProxy(float(x)))
            else:
                raise ValueError(
                    f"Expected UGen, UGenScalar, UGenVector, or float, got {type(x).__name__}: {x!r}"
                )
        self._values = tuple(values_)

    @overload
    def __getitem__(self, i: int) -> UGenOperable: ...
    @overload
    def __getitem__(self, i: slice) -> "UGenVector": ...
    def __getitem__(self, i: int | slice) -> "UGenOperable | UGenVector":
        if isinstance(i, int):
            return self._values[i]
        return UGenVector(*self._values[i])

    def __iter__(self) -> Iterator[UGenOperable]:
        yield from self._values

    def __len__(self) -> int:
        return len(self._values)

    def __repr__(self) -> str:
        return f"<UGenVector([{', '.join(repr(x) for x in self)}])>"


class SynthDefError(Exception):
    pass


# Thread-local storage for active builders
_local = threading.local()
_local._active_builders = []


class UGen(UGenOperable, SequenceABC["UGenOperable"]):
    """Base class for all unit generators."""

    __slots__ = (
        "_calculation_rate",
        "_special_index",
        "_inputs",
        "_input_keys",
        "_uuid",
        "_values",
    )

    _channel_count = 1
    _has_done_flag = False
    _is_output = False
    _is_pure = False
    _is_width_first = False
    _ordered_keys: tuple[str, ...] = ()
    _unexpanded_keys: frozenset[str] = frozenset()

    def __init__(
        self,
        *,
        calculation_rate: CalculationRate = CalculationRate.SCALAR,
        special_index: SupportsInt = 0,
        **kwargs: UGenRecursiveInput | None,
    ) -> None:
        calculation_rate, kwargs = self._postprocess_kwargs(
            calculation_rate=calculation_rate, **kwargs
        )
        self._calculation_rate = CalculationRate.from_expr(calculation_rate)
        self._special_index = int(special_index)
        input_keys: list[str | tuple[str, int]] = []
        inputs: list[OutputProxy | float] = []
        for key in self._ordered_keys:
            value = kwargs.pop(key, None)
            if value is None:
                continue
            if isinstance(value, UGenSerializable):
                serialized = value.serialize()
                if any(isinstance(x, UGenVector) for x in serialized):
                    raise ValueError(
                        f"Unexpected UGenVector in serialized input '{key}'"
                    )
                value = cast(SequenceABC[SupportsFloat | UGenScalar], serialized)
            if isinstance(value, SequenceABC) and not isinstance(
                value, (str, UGenScalar)
            ):
                if key not in self._unexpanded_keys:
                    raise ValueError(
                        f"Sequence input for '{key}' requires unexpanded=True in param()"
                    )
                iterator: Iterable[tuple[int | None, Any]] = (
                    (i, v) for i, v in enumerate(value)
                )
            else:
                iterator = ((None, v) for v in [value])
            for i, x in iterator:
                if isinstance(x, ConstantProxy):
                    inputs.append(float(x.value))
                elif isinstance(x, OutputProxy):
                    inputs.append(x)
                elif isinstance(x, SupportsFloat):
                    inputs.append(float(x))
                else:
                    raise ValueError(
                        f"Invalid input type for '{key}': expected float or UGenScalar, got {type(x).__name__}"
                    )
                input_keys.append((key, i) if i is not None else key)
        if kwargs:
            raise ValueError(
                f"{type(self).__name__} received unknown parameters: {', '.join(kwargs)}"
            )
        self._inputs = tuple(inputs)
        self._input_keys = tuple(input_keys)
        self._uuid: uuid.UUID | None = None
        if hasattr(_local, "_active_builders") and _local._active_builders:
            builder = _local._active_builders[-1]
            self._uuid = builder._uuid
            builder._add_ugen(self)
        for input_ in self._inputs:
            if isinstance(input_, OutputProxy) and input_.ugen._uuid != self._uuid:
                raise SynthDefError("UGen input in different scope")
        self._values = tuple(
            OutputProxy(ugen=self, index=i)
            for i in range(getattr(self, "_channel_count", 1))
        )

    @overload
    def __getitem__(self, i: int) -> UGenOperable: ...
    @overload
    def __getitem__(self, i: slice) -> UGenVector: ...
    def __getitem__(self, i: int | slice) -> UGenOperable | UGenVector:
        if isinstance(i, int):
            return self._values[i]
        return UGenVector(*self._values[i])

    def __iter__(self) -> Iterator[UGenOperable]:
        yield from self._values

    def __len__(self) -> int:
        return self._channel_count

    def __repr__(self) -> str:
        return f"<{type(self).__name__}.{self.calculation_rate.token}()>"

    def _eliminate(
        self, sort_bundles: dict["UGen", "SynthDefBuilder.SortBundle"]
    ) -> None:
        if not (sort_bundle := sort_bundles.get(self)) or sort_bundle.descendants:
            return
        del sort_bundles[self]
        for antecedent in tuple(sort_bundle.antecedents):
            if not (antecedent_bundle := sort_bundles.get(antecedent)):
                continue
            antecedent_bundle.descendants.remove(self)
            antecedent._optimize(sort_bundles)

    @classmethod
    def _expand_params(
        cls,
        params: dict[str, UGenRecursiveInput],
        unexpanded_keys: Iterable[str] | None = None,
    ) -> UGenRecursiveParams:
        unexpanded_keys_ = set(unexpanded_keys or ())
        size = 0
        for key, value in params.items():
            if isinstance(value, UGenSerializable):
                params[key] = value = value.serialize()
            if isinstance(value, (SupportsFloat, UGenScalar)):
                continue
            elif isinstance(value, SequenceABC) and not isinstance(value, str):
                if key in unexpanded_keys_:
                    if isinstance(value, SequenceABC) and any(
                        (
                            isinstance(x, SequenceABC)
                            and not isinstance(x, (SupportsFloat, UGenScalar, str))
                        )
                        for x in value
                    ):
                        size = max(size, len(value))
                    else:
                        continue
                else:
                    size = max(size, len(value))
        if not size:
            return cast(dict[str, Union[UGenScalarInput, UGenVectorInput]], params)
        results = []
        for i in range(size):
            new_params: dict[str, UGenRecursiveInput] = {}
            for key, value in params.items():
                if isinstance(value, UGenSerializable):
                    value = value.serialize()
                if isinstance(value, (SupportsFloat, UGenScalar)):
                    new_params[key] = value
                elif isinstance(value, SequenceABC) and not isinstance(value, str):
                    if key in unexpanded_keys_:
                        if isinstance(value, SequenceABC) and all(
                            isinstance(x, (SupportsFloat, UGenScalar)) for x in value
                        ):
                            new_params[key] = value
                        else:
                            new_params[key] = value[i % len(value)]
                    else:
                        new_params[key] = value[i % len(value)]
            results.append(
                cls._expand_params(new_params, unexpanded_keys=unexpanded_keys)
            )
        return results

    @classmethod
    def _new_expanded(
        cls,
        *,
        calculation_rate: CalculationRate | None,
        special_index: int = 0,
        **kwargs: UGenRecursiveInput | None,
    ) -> UGenOperable:
        def recurse(all_expanded_params: UGenRecursiveParams) -> UGenOperable:
            if (
                not isinstance(all_expanded_params, dict)
                and len(all_expanded_params) == 1
            ):
                all_expanded_params = all_expanded_params[0]
            if isinstance(all_expanded_params, dict):
                return cls._new_single(
                    calculation_rate=calculation_rate,
                    special_index=special_index,
                    **all_expanded_params,
                )
            return UGenVector(
                *(recurse(expanded_params) for expanded_params in all_expanded_params)
            )

        filtered = {k: v for k, v in kwargs.items() if v is not None}
        return recurse(
            cls._expand_params(filtered, unexpanded_keys=cls._unexpanded_keys)
        )

    @classmethod
    def _new_single(
        cls,
        *,
        calculation_rate: CalculationRate | None = None,
        special_index: SupportsInt = 0,
        **kwargs: UGenRecursiveInput | None,
    ) -> UGenOperable:
        ugen = cls(
            calculation_rate=CalculationRate.from_expr(calculation_rate),
            special_index=special_index,
            **kwargs,
        )
        if len(ugen) == 1:
            return ugen[0]
        return ugen

    def _optimize(
        self, sort_bundles: dict["UGen", "SynthDefBuilder.SortBundle"]
    ) -> None:
        if not self._is_pure:
            return
        self._eliminate(sort_bundles)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs: UGenRecursiveInput | None,
    ) -> tuple[CalculationRate, dict[str, UGenRecursiveInput | None]]:
        return calculation_rate, kwargs

    @property
    def calculation_rate(self) -> CalculationRate:
        return self._calculation_rate

    @property
    def has_done_flag(self) -> bool:
        return self._has_done_flag

    @property
    def inputs(self) -> tuple[OutputProxy | float, ...]:
        return tuple(self._inputs)

    @property
    def special_index(self) -> int:
        return self._special_index


# ---------------------------------------------------------------------------
# Operator UGens
# ---------------------------------------------------------------------------


class UnaryOpUGen(UGen):
    _ordered_keys = ("source",)
    _is_pure = True

    def __init__(
        self,
        *,
        calculation_rate: CalculationRate,
        source: UGenScalarInput,
        special_index: SupportsInt = 0,
    ) -> None:
        super().__init__(
            calculation_rate=calculation_rate,
            source=source,
            special_index=special_index,
        )

    def __repr__(self) -> str:
        return f"<UnaryOpUGen.{self.calculation_rate.token}({self.operator.name})>"

    @property
    def operator(self) -> UnaryOperator:
        return UnaryOperator(self.special_index)


class BinaryOpUGen(UGen):
    _ordered_keys = ("left", "right")
    _is_pure = True

    def __init__(
        self,
        *,
        calculation_rate: CalculationRate,
        left: UGenScalarInput,
        right: UGenScalarInput,
        special_index: SupportsInt = 0,
    ) -> None:
        super().__init__(
            calculation_rate=calculation_rate,
            left=left,
            right=right,
            special_index=special_index,
        )

    def __repr__(self) -> str:
        return f"<BinaryOpUGen.{self.calculation_rate.token}({self.operator.name})>"

    @classmethod
    def _new_single(
        cls,
        *,
        calculation_rate: CalculationRate | None = None,
        special_index: SupportsInt = 0,
        **kwargs: UGenRecursiveInput | None,
    ) -> UGenOperable:
        def process(
            left: UGenScalar | float,
            right: UGenScalar | float,
        ) -> UGenOperable | float:
            if special_index == BinaryOperator.MULTIPLICATION:
                if left == 0 or right == 0:
                    return ConstantProxy(0)
                if left == 1:
                    return right
                if left == -1:
                    return -right
                if right == 1:
                    return left
                if right == -1:
                    return -left
            if special_index == BinaryOperator.ADDITION:
                if left == 0:
                    return right
                if right == 0:
                    return left
            if special_index == BinaryOperator.SUBTRACTION:
                if left == 0:
                    return -right
                if right == 0:
                    return left
            if special_index == BinaryOperator.FLOAT_DIVISION:
                if right == 1:
                    return left
                if right == -1:
                    return -left
            return cls(
                calculation_rate=max(
                    [
                        CalculationRate.from_expr(left),
                        CalculationRate.from_expr(right),
                    ]
                ),
                special_index=special_index,
                left=left,
                right=right,
            )[0]

        left = kwargs["left"]
        right = kwargs["right"]
        if not isinstance(left, (SupportsFloat, UGenScalar)):
            raise ValueError(
                f"Left operand must be float or UGenScalar, got {type(left).__name__}"
            )
        if not isinstance(right, (SupportsFloat, UGenScalar)):
            raise ValueError(
                f"Right operand must be float or UGenScalar, got {type(right).__name__}"
            )
        result = process(
            float(left) if isinstance(left, SupportsFloat) else left,
            float(right) if isinstance(right, SupportsFloat) else right,
        )
        if isinstance(result, SupportsFloat) and not isinstance(result, UGenOperable):
            return ConstantProxy(result)
        if not isinstance(result, UGenOperable):
            return ConstantProxy(float(result))
        return result

    @property
    def operator(self) -> BinaryOperator:
        return BinaryOperator(self.special_index)


# ---------------------------------------------------------------------------
# Parameter / Control
# ---------------------------------------------------------------------------


class Parameter(UGen):
    def __init__(
        self,
        *,
        name: str | None = None,
        value: float | SequenceABC[float],
        rate: ParameterRate | None = ParameterRate.CONTROL,
        lag: float | None = None,
    ) -> None:
        if isinstance(value, SupportsFloat):
            self.value: tuple[float, ...] = (float(value),)
        else:
            self.value = tuple(float(x) for x in value)
        self.name = name
        self.lag = lag
        self.rate = ParameterRate.from_expr(rate)
        self._channel_count = len(self.value)
        super().__init__(calculation_rate=CalculationRate.from_expr(self.rate))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return (type(self), self.name, self.value, self.rate, self.lag) == (
            type(other),
            other.name,
            other.value,
            other.rate,
            other.lag,
        )

    def __hash__(self) -> int:
        return hash((type(self), self.name, self.value, self.rate, self.lag))

    def __repr__(self) -> str:
        return f"<Parameter.{self.calculation_rate.token}({self.name})>"


class Control(UGen):
    def __init__(
        self,
        *,
        parameters: SequenceABC[Parameter],
        calculation_rate: CalculationRate,
        special_index: int = 0,
    ) -> None:
        self._parameters = tuple(parameters)
        self._channel_count = sum(len(parameter) for parameter in self._parameters)
        super().__init__(
            calculation_rate=calculation_rate,
            special_index=special_index,
        )

    @property
    def parameters(self) -> SequenceABC[Parameter]:
        return self._parameters


class AudioControl(Control):
    pass


class LagControl(Control):
    _ordered_keys = ("lags",)
    _unexpanded_keys = frozenset(["lags"])

    def __init__(
        self,
        *,
        parameters: SequenceABC[Parameter],
        calculation_rate: CalculationRate,
        special_index: int = 0,
    ) -> None:
        self._parameters = tuple(parameters)
        self._channel_count = sum(len(parameter) for parameter in self._parameters)
        lags = []
        for parameter in parameters:
            lags.extend([parameter.lag or 0.0] * len(parameter))
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            lags=lags,
            special_index=special_index,
        )


class TrigControl(Control):
    pass


# ---------------------------------------------------------------------------
# SynthDef
# ---------------------------------------------------------------------------


class SynthDef:
    def __init__(self, ugens: SequenceABC[UGen], name: str | None = None) -> None:
        if not ugens:
            raise SynthDefError("No UGens provided")
        self._ugens = tuple(ugens)
        self._name = name
        constants: list[float] = []
        for ugen in ugens:
            for input_ in ugen.inputs:
                if isinstance(input_, float) and input_ not in constants:
                    constants.append(input_)
        self._constants = tuple(constants)
        self._controls: tuple[Control, ...] = tuple(
            ugen for ugen in ugens if isinstance(ugen, Control)
        )
        self._parameters: dict[str, tuple[Parameter, int]] = (
            self._collect_indexed_parameters(self._controls)
        )
        self._compiled_graph = _compile_ugen_graph(self)

    def __hash__(self) -> int:
        return hash((type(self), self._name, self._compiled_graph))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return (self._name, self._compiled_graph) == (
            other._name,
            other._compiled_graph,
        )

    def __repr__(self) -> str:
        return f"<SynthDef: {self.effective_name}>"

    def _collect_indexed_parameters(
        self, controls: SequenceABC[Control]
    ) -> dict[str, tuple[Parameter, int]]:
        mapping: dict[str, tuple[Parameter, int]] = {}
        for control in controls:
            index = control.special_index
            for parameter in control.parameters:
                if parameter.name is None:
                    raise ValueError(f"Parameter has no name: {parameter!r}")
                mapping[parameter.name] = (parameter, index)
                index += len(parameter)
        return mapping

    def compile(self, use_anonymous_name: bool = False) -> bytes:
        return compile_synthdefs(self, use_anonymous_names=use_anonymous_name)

    @property
    def anonymous_name(self) -> str:
        return hashlib.md5(self._compiled_graph).hexdigest()

    @property
    def constants(self) -> SequenceABC[float]:
        return self._constants

    @property
    def controls(self) -> SequenceABC[Control]:
        return self._controls

    @property
    def effective_name(self) -> str:
        return self.name or self.anonymous_name

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def parameters(self) -> dict[str, tuple[Parameter, int]]:
        return dict(self._parameters)

    @property
    def ugens(self) -> SequenceABC[UGen]:
        return self._ugens

    # -- High-level convenience methods ----------------------------------------

    def send(self, server: "Any") -> None:
        """Send this SynthDef to a running server."""
        server.send_synthdef(self)

    def play(
        self,
        server: "Any",
        target: int = 1,
        action: int = 0,
        **params: float,
    ) -> int:
        """Send this SynthDef (if needed) and create a synth. Returns node ID."""
        self.send(server)
        result: int = server.synth(
            self.effective_name, target=target, action=action, **params
        )
        return result

    def dump_ugens(self) -> str:
        """Return a human-readable representation of the UGen graph."""
        lines: list[str] = [f"SynthDef: {self.effective_name}"]
        for i, u in enumerate(self._ugens):
            type_name = type(u).__name__
            rate = u.calculation_rate.token
            if isinstance(u, Control):
                param_names = [p.name or "?" for p in u.parameters]
                line = f"  {i}: {type_name}.{rate} - [{', '.join(param_names)}]"
            else:
                parts: list[str] = []
                for input_, key in zip(u._inputs, u._input_keys):
                    if isinstance(key, tuple):
                        key_str = key[0]
                    else:
                        key_str = key
                    if isinstance(input_, OutputProxy):
                        src_name = type(input_.ugen).__name__
                        parts.append(f"{key_str}: {src_name}[{input_.index}]")
                    else:
                        parts.append(f"{key_str}: {input_}")
                if isinstance(u, BinaryOpUGen):
                    op_name = u.operator.name
                    line = f"  {i}: {type_name}.{rate}({op_name}, {', '.join(parts)})"
                elif isinstance(u, UnaryOpUGen):
                    op_name = u.operator.name
                    line = f"  {i}: {type_name}.{rate}({op_name}, {', '.join(parts)})"
                else:
                    line = f"  {i}: {type_name}.{rate}({', '.join(parts)})"
                if len(u) > 1:
                    line += f" -> {len(u)} outputs"
            lines.append(line)
        return "\n".join(lines)


# ---------------------------------------------------------------------------
# SynthDefBuilder
# ---------------------------------------------------------------------------


class SynthDefBuilder:
    class SortBundle(NamedTuple):
        ugen: UGen
        width_first_antecedents: tuple[UGen, ...]
        antecedents: list[UGen]
        descendants: list[UGen]

    def __init__(self, **kwargs: Parameter | SequenceABC[float] | float) -> None:
        self._building = False
        self._parameters: dict[str, Parameter] = {}
        self._ugens: list[UGen] = []
        self._uuid = uuid.uuid4()
        if not hasattr(_local, "_active_builders"):
            _local._active_builders = []
        for key, value in kwargs.items():
            if isinstance(value, Parameter):
                self.add_parameter(
                    lag=value.lag, name=key, value=value.value, rate=value.rate
                )
            else:
                self.add_parameter(name=key, value=value)

    def __enter__(self) -> "SynthDefBuilder":
        if not hasattr(_local, "_active_builders"):
            _local._active_builders = []
        _local._active_builders.append(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        _local._active_builders.pop()

    def __getitem__(self, item: str) -> OutputProxy | Parameter:
        """Look up a parameter by name.

        Returns an OutputProxy for scalar parameters, or the Parameter
        itself for multi-channel parameters (sequence values).

        Raises:
            KeyError: If the parameter name does not exist.
        """
        parameter = self._parameters[item]
        if len(parameter) == 1:
            return cast(OutputProxy, parameter[0])
        return parameter

    def _add_ugen(self, ugen: UGen) -> None:
        if ugen._uuid != self._uuid:
            raise SynthDefError("UGen input in different scope")
        if not self._building:
            self._ugens.append(ugen)

    def _build_control_mapping(
        self, parameters: SequenceABC[Parameter]
    ) -> tuple[list[Control], dict[OutputProxy, OutputProxy]]:
        parameter_mapping: dict[ParameterRate, list[Parameter]] = {}
        for parameter in parameters:
            parameter_mapping.setdefault(parameter.rate, []).append(parameter)
        for filtered_parameters in parameter_mapping.values():
            filtered_parameters.sort(key=lambda x: x.name or "")
        controls: list[Control] = []
        control_mapping: dict[OutputProxy, OutputProxy] = {}
        starting_control_index = 0
        for parameter_rate in sorted(ParameterRate):
            filtered_parameters = parameter_mapping.get(parameter_rate, [])
            if not filtered_parameters:
                continue
            if parameter_rate == ParameterRate.SCALAR:
                control = Control(
                    calculation_rate=CalculationRate.SCALAR,
                    parameters=filtered_parameters,
                    special_index=starting_control_index,
                )
            elif parameter_rate == ParameterRate.TRIGGER:
                control = TrigControl(
                    calculation_rate=CalculationRate.CONTROL,
                    parameters=filtered_parameters,
                    special_index=starting_control_index,
                )
            elif parameter_rate == ParameterRate.AUDIO:
                control = AudioControl(
                    calculation_rate=CalculationRate.AUDIO,
                    parameters=filtered_parameters,
                    special_index=starting_control_index,
                )
            elif any(parameter.lag for parameter in filtered_parameters):
                control = LagControl(
                    calculation_rate=CalculationRate.CONTROL,
                    parameters=filtered_parameters,
                    special_index=starting_control_index,
                )
            else:
                control = Control(
                    calculation_rate=CalculationRate.CONTROL,
                    parameters=filtered_parameters,
                    special_index=starting_control_index,
                )
            controls.append(control)
            output_index = 0
            for parameter in filtered_parameters:
                for output in parameter:
                    control_mapping[cast(OutputProxy, output)] = cast(
                        OutputProxy, control[output_index]
                    )
                    output_index += 1
                    starting_control_index += 1
        return controls, control_mapping

    def _initiate_topological_sort(
        self, ugens: list[UGen]
    ) -> dict[UGen, "SynthDefBuilder.SortBundle"]:
        sort_bundles: dict[UGen, SynthDefBuilder.SortBundle] = {}
        width_first_antecedents: list[UGen] = []
        for ugen in ugens:
            sort_bundles[ugen] = self.SortBundle(
                antecedents=[],
                descendants=[],
                ugen=ugen,
                width_first_antecedents=tuple(width_first_antecedents),
            )
            if ugen._is_width_first:
                width_first_antecedents.append(ugen)
        for ugen, sort_bundle in sort_bundles.items():
            for input_ in ugen.inputs:
                if not isinstance(input_, OutputProxy):
                    continue
                if input_.ugen not in sort_bundle.antecedents:
                    sort_bundle.antecedents.append(input_.ugen)
                if (
                    ugen
                    not in (input_sort_bundle := sort_bundles[input_.ugen]).descendants
                ):
                    input_sort_bundle.descendants.append(ugen)
            for antecedent in sort_bundle.width_first_antecedents:
                if antecedent not in sort_bundle.antecedents:
                    sort_bundle.antecedents.append(antecedent)
                if (
                    ugen
                    not in (input_sort_bundle := sort_bundles[antecedent]).descendants
                ):
                    input_sort_bundle.descendants.append(ugen)
            sort_bundle.descendants[:] = sorted(
                sort_bundles[ugen].descendants,
                key=lambda x: ugens.index(ugen),
            )
        return sort_bundles

    def _cleanup_local_bufs(self, ugens: list[UGen]) -> list[UGen]:
        """Insert MaxLocalBufs for any LocalBuf UGens in the graph.

        SuperCollider requires a MaxLocalBufs UGen to be present whenever
        LocalBuf UGens are used (e.g. by FFT). This pass counts LocalBuf
        instances, inserts a MaxLocalBufs just before the first one, and
        wires them together.
        """
        from .ugens.bufio import LocalBuf, MaxLocalBufs

        filtered: list[UGen] = []
        local_bufs: list[UGen] = []
        for ugen in ugens:
            if isinstance(ugen, MaxLocalBufs):
                continue  # remove existing MaxLocalBufs; we'll rebuild
            if isinstance(ugen, LocalBuf):
                local_bufs.append(ugen)
            filtered.append(ugen)
        if local_bufs:
            max_local_bufs = cast(
                OutputProxy,
                MaxLocalBufs.ir(maximum=len(local_bufs)),  # type: ignore[attr-defined]
            )
            for local_buf in local_bufs:
                inputs: list[OutputProxy | float] = list(local_buf._inputs[:2])
                inputs.append(max_local_bufs)
                local_buf._inputs = tuple(inputs)
            index = filtered.index(local_bufs[0])
            filtered[index:index] = [max_local_bufs.ugen]
        return filtered

    def _optimize(self, ugens: list[UGen]) -> list[UGen]:
        sort_bundles = self._initiate_topological_sort(ugens)
        for ugen in ugens:
            ugen._optimize(sort_bundles)
        return list(sort_bundles)

    def _remap_controls(
        self,
        ugens: list[UGen],
        control_mapping: dict[OutputProxy, OutputProxy],
    ) -> list[UGen]:
        for ugen in ugens:
            ugen._inputs = tuple(
                (
                    control_mapping.get(input_, input_)
                    if isinstance(input_, OutputProxy)
                    else input_
                )
                for input_ in ugen._inputs
            )
        return ugens

    def _sort_topologically(self, ugens: list[UGen]) -> list[UGen]:
        sort_bundles = self._initiate_topological_sort(ugens)
        available_ugens: list[UGen] = []
        output_stack: list[UGen] = []
        for ugen in reversed(ugens):
            if not sort_bundles[ugen].antecedents and ugen not in available_ugens:
                available_ugens.append(ugen)
        while available_ugens:
            available_ugen = available_ugens.pop()
            for descendant in reversed(sort_bundles[available_ugen].descendants):
                descendant_sort_bundle = sort_bundles[descendant]
                descendant_sort_bundle.antecedents.remove(available_ugen)
                if (
                    not descendant_sort_bundle.antecedents
                    and descendant_sort_bundle.ugen not in available_ugens
                ):
                    available_ugens.append(descendant_sort_bundle.ugen)
            output_stack.append(available_ugen)
        return output_stack

    def add_parameter(
        self,
        *,
        name: str,
        value: float | SequenceABC[float],
        rate: ParameterRate | None = ParameterRate.CONTROL,
        lag: float | None = None,
    ) -> OutputProxy | Parameter:
        """Add a named parameter to the SynthDef.

        Args:
            name: Parameter name (must be unique within this builder).
            value: Default value. A sequence creates a multi-channel parameter.
            rate: Parameter rate (CONTROL, AUDIO, TRIGGER, or SCALAR).
            lag: Lag time in seconds. Produces a LagControl when set.

        Returns:
            OutputProxy for scalar parameters, or the Parameter for
            multi-channel parameters.

        Raises:
            ValueError: If a parameter with this name already exists.
        """
        if name in self._parameters:
            raise ValueError(f"Duplicate parameter name: '{name}'")
        with self:
            parameter = Parameter(
                lag=lag, name=name, rate=ParameterRate.from_expr(rate), value=value
            )
        self._parameters[name] = parameter
        if len(parameter) == 1:
            return cast(OutputProxy, parameter[0])
        return parameter

    def build(self, name: str | None = None, optimize: bool = True) -> SynthDef:
        """Compile the UGen graph into a SynthDef.

        Performs control mapping, topological sorting, and optional dead
        code elimination. The builder can be reused after calling build().

        Args:
            name: SynthDef name. If None, an anonymous MD5 hash is used.
            optimize: If True, eliminate unused pure UGens from the graph.

        Returns:
            A compiled SynthDef ready for sending to a server.
        """
        try:
            self._building = True
            with self:
                ugens: list[UGen] = copy.deepcopy(self._ugens)
                parameters: list[Parameter] = sorted(
                    [x for x in ugens if isinstance(x, Parameter)],
                    key=lambda x: x.name or "",
                )
                ugens = [x for x in ugens if not isinstance(x, Parameter)]
                controls, control_mapping = self._build_control_mapping(parameters)
                ugens = controls + ugens
                ugens = self._remap_controls(ugens, control_mapping)
                ugens = self._cleanup_local_bufs(ugens)
                ugens = self._sort_topologically(ugens)
                if optimize:
                    ugens = self._optimize(ugens)
        finally:
            self._building = False
        return SynthDef(ugens, name=name)


from .compiler import _compile_ugen_graph, compile_synthdefs  # noqa: E402


# ---------------------------------------------------------------------------
# @synthdef decorator
# ---------------------------------------------------------------------------


def synthdef(*args: str | tuple[str, float]) -> Callable[..., SynthDef]:
    """Decorator for constructing SynthDefs from functions.

    Parameter rates and lags can be specified positionally::

        @synthdef("ar", ("kr", 0.5))
        def my_synth(freq=440, amp=0.1):
            ...

    Without arguments, all parameters default to control rate::

        @synthdef()
        def my_synth(freq=440, amp=0.1):
            ...
    """
    import inspect

    def inner(func: Callable[..., Any]) -> SynthDef:
        signature = inspect.signature(func)
        builder = SynthDefBuilder()
        kwargs: dict[str, OutputProxy | Parameter] = {}
        for i, (name, parameter) in enumerate(signature.parameters.items()):
            rate = ParameterRate.CONTROL
            lag = None
            try:
                arg_i = args[i]
                if isinstance(arg_i, str):
                    rate = ParameterRate.from_expr(arg_i)
                else:
                    rate_expr, lag = arg_i
                    rate = ParameterRate.from_expr(rate_expr)
            except (IndexError, TypeError):
                pass
            value = parameter.default
            if value is inspect._empty:
                value = 0.0
            kwargs[name] = builder.add_parameter(
                name=name, lag=lag, rate=rate, value=value
            )
        with builder:
            func(**kwargs)
        return builder.build(name=func.__name__)

    return inner



"""Minimal standalone OSC message/bundle encode/decode."""

import contextlib
import datetime
import socket
import struct
import time
from collections.abc import Iterator, Sequence as SequenceABC
from typing import Any, Union

try:
    from . import _osc as _osc_native  # type: ignore[attr-defined]
except ImportError:
    _osc_native = None

BUNDLE_PREFIX = b"#bundle\x00"
IMMEDIATELY = struct.pack(">Q", 1)
NTP_TIMESTAMP_TO_SECONDS = 1.0 / 2.0**32.0
SECONDS_TO_NTP_TIMESTAMP = 2.0**32.0
SYSTEM_EPOCH = datetime.date(*time.gmtime(0)[0:3])
NTP_EPOCH = datetime.date(1900, 1, 1)
NTP_DELTA = (SYSTEM_EPOCH - NTP_EPOCH).days * 24 * 3600


OscArgument = Union[
    "OscBundle",
    "OscMessage",
    SequenceABC["OscArgument"],
    bool,
    bytes,
    float,
    str,
    None,
]


def _group_by_count(iterable: SequenceABC[Any], count: int) -> Iterator[list[Any]]:
    """Split an iterable into chunks of ``count`` items."""
    iterator = iter(iterable)
    result: list[Any] = []
    for item in iterator:
        result.append(item)
        if len(result) == count:
            yield result
            result = []
    if result:
        yield result


def format_datagram(datagram: bytes | bytearray) -> str:
    result: list[str] = ["size {}".format(len(datagram))]
    index = 0
    while index < len(datagram):
        hex_blocks = []
        ascii_block = ""
        for chunk in _group_by_count(datagram[index : index + 16], 4):
            hex_block = []
            for byte in chunk:
                if 31 < int(byte) < 127:
                    char = chr(int(byte))
                else:
                    char = "."
                ascii_block += char
                hexed = hex(byte)[2:].zfill(2)
                hex_block.append(hexed)
            hex_blocks.append(" ".join(hex_block))
        line = "{: >4}   ".format(index)
        line += "{: <53}".format("  ".join(hex_blocks))
        line += "|{}|".format(ascii_block)
        result.append(line)
        index += 16
    return "\n".join(result)


class OscMessage:
    """An OSC message."""

    def __init__(self, address: int | str, *contents: OscArgument) -> None:
        if not isinstance(address, (str, int)):
            raise ValueError(f"address must be int or str, got {address}")
        self.address = address
        self.contents = tuple(contents)

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other):
            return False
        assert isinstance(other, OscMessage)
        if self.address != other.address:
            return False
        if self.contents != other.contents:
            return False
        return True

    def __repr__(self) -> str:
        return "{}({})".format(
            type(self).__name__,
            ", ".join(repr(_) for _ in [self.address, *self.contents]),
        )

    def __str__(self) -> str:
        return format_datagram(bytearray(self.to_datagram()))

    @staticmethod
    def _decode_blob(data: bytes) -> tuple[bytes, bytes]:
        actual_length, remainder = struct.unpack(">I", data[:4])[0], data[4:]
        padded_length = actual_length
        if actual_length % 4 != 0:
            padded_length = (actual_length // 4 + 1) * 4
        return remainder[:padded_length][:actual_length], remainder[padded_length:]

    @staticmethod
    def _decode_string(data: bytes) -> tuple[str, bytes]:
        actual_length = data.index(b"\x00")
        padded_length = (actual_length // 4 + 1) * 4
        return str(data[:actual_length], "ascii"), data[padded_length:]

    @staticmethod
    def _encode_string(value: str) -> bytes:
        result = bytes(value + "\x00", "ascii")
        if len(result) % 4 != 0:
            width = (len(result) // 4 + 1) * 4
            result = result.ljust(width, b"\x00")
        return result

    @staticmethod
    def _encode_blob(value: bytes) -> bytes:
        result = bytes(struct.pack(">I", len(value)) + value)
        if len(result) % 4 != 0:
            width = (len(result) // 4 + 1) * 4
            result = result.ljust(width, b"\x00")
        return result

    @classmethod
    def _encode_value(cls, value: OscArgument) -> tuple[str, bytes]:
        type_tags, encoded_value = "", b""
        if isinstance(value, (OscBundle, OscMessage)):
            type_tags += "b"
            encoded_value = cls._encode_blob(value.to_datagram())
        elif isinstance(value, (bytearray, bytes)):
            type_tags += "b"
            encoded_value = cls._encode_blob(bytes(value))
        elif isinstance(value, str):
            type_tags += "s"
            encoded_value = cls._encode_string(value)
        elif isinstance(value, bool):
            type_tags += "T" if value else "F"
        elif isinstance(value, float):
            type_tags += "f"
            encoded_value += struct.pack(">f", value)
        elif isinstance(value, int):
            type_tags += "i"
            encoded_value += struct.pack(">i", value)
        elif value is None:
            type_tags += "N"
        elif isinstance(value, SequenceABC):
            type_tags += "["
            for sub_value in value:
                sub_type_tags, sub_encoded_value = cls._encode_value(sub_value)
                type_tags += sub_type_tags
                encoded_value += sub_encoded_value
            type_tags += "]"
        else:
            message = "Cannot encode {!r}".format(value)
            raise TypeError(message)
        return type_tags, encoded_value

    def to_datagram(self) -> bytes:
        if _osc_native is not None:
            if isinstance(self.address, str):
                return bytes(_osc_native.encode_message(self.address, *self.contents))
            else:
                return bytes(
                    _osc_native.encode_message_int(self.address, *self.contents)
                )
        # Fallback: pure Python
        if isinstance(self.address, str):
            encoded_address = self._encode_string(self.address)
        else:
            encoded_address = struct.pack(">i", self.address)
        encoded_type_tags = ","
        encoded_contents = b""
        for value in self.contents or ():
            type_tags, encoded_value = self._encode_value(value)
            encoded_type_tags += type_tags
            encoded_contents += encoded_value
        return (
            encoded_address + self._encode_string(encoded_type_tags) + encoded_contents
        )

    @classmethod
    def from_datagram(cls, datagram: bytes) -> "OscMessage":
        if _osc_native is not None:
            address, contents = _osc_native.decode_message(datagram)
            return cls(address, *contents)
        # Fallback: pure Python
        remainder = datagram
        address, remainder = cls._decode_string(remainder)
        type_tags, remainder = cls._decode_string(remainder)
        contents_list: list[OscArgument] = []
        array_stack: list[list[OscArgument]] = [contents_list]
        for type_tag in type_tags[1:]:
            if type_tag == "i":
                value, remainder = struct.unpack(">i", remainder[:4])[0], remainder[4:]
                array_stack[-1].append(value)
            elif type_tag == "f":
                value, remainder = struct.unpack(">f", remainder[:4])[0], remainder[4:]
                array_stack[-1].append(value)
            elif type_tag == "d":
                value, remainder = struct.unpack(">d", remainder[:8])[0], remainder[8:]
                array_stack[-1].append(value)
            elif type_tag == "s":
                value, remainder = cls._decode_string(remainder)
                array_stack[-1].append(value)
            elif type_tag == "b":
                value, remainder = cls._decode_blob(remainder)
                for class_ in (OscBundle, OscMessage):
                    try:
                        value = class_.from_datagram(value)
                        break
                    except (ValueError, IndexError, struct.error):
                        pass
                array_stack[-1].append(value)
            elif type_tag == "T":
                array_stack[-1].append(True)
            elif type_tag == "F":
                array_stack[-1].append(False)
            elif type_tag == "N":
                array_stack[-1].append(None)
            elif type_tag == "[":
                array: list[OscArgument] = []
                array_stack[-1].append(array)
                array_stack.append(array)
            elif type_tag == "]":
                array_stack.pop()
            else:
                raise RuntimeError(f"Unable to parse type {type_tag!r}")
        return cls(address, *contents_list)

    def to_list(self) -> list[Any]:
        result: list[Any] = [self.address]
        for x in self.contents:
            if isinstance(x, (OscMessage, OscBundle)):
                result.append(x.to_list())
            else:
                result.append(x)
        return result

    def to_osc(self) -> "OscMessage":
        return self


class OscBundle:
    """An OSC bundle."""

    def __init__(
        self,
        timestamp: float | None = None,
        *,
        contents: SequenceABC["OscBundle | OscMessage"],
    ) -> None:
        prototype = (OscMessage, type(self))
        self.timestamp = timestamp
        contents = contents or ()
        for x in contents or ():
            if not isinstance(x, prototype):
                raise ValueError(contents)
        self.contents = tuple(contents)

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other):
            return False
        assert isinstance(other, OscBundle)
        if self.timestamp != other.timestamp:
            return False
        if self.contents != other.contents:
            return False
        return True

    def __repr__(self) -> str:
        parts = ["{}(".format(type(self).__name__)]
        if self.timestamp is not None:
            parts.append(f"timestamp={self.timestamp}")
            if self.contents:
                parts.append(", ")
        if self.contents:
            parts.append(f"contents={list(self.contents)!r}")
        parts.append(")")
        return "".join(parts)

    def __str__(self) -> str:
        return format_datagram(bytearray(self.to_datagram()))

    @staticmethod
    def _decode_date(data: bytes) -> tuple[float | None, bytes]:
        data, remainder = data[:8], data[8:]
        if data == IMMEDIATELY:
            return None, remainder
        date = (struct.unpack(">Q", data)[0] / SECONDS_TO_NTP_TIMESTAMP) - NTP_DELTA
        return date, remainder

    @staticmethod
    def _encode_date(seconds: float | None, realtime: bool = True) -> bytes:
        if seconds is None:
            return IMMEDIATELY
        if realtime:
            seconds = seconds + NTP_DELTA
        if seconds >= 4294967296:  # 2**32
            seconds = seconds % 4294967296
        return struct.pack(">Q", int(seconds * SECONDS_TO_NTP_TIMESTAMP))

    @classmethod
    def from_datagram(cls, datagram: bytes) -> "OscBundle":
        if _osc_native is not None:
            if not datagram.startswith(BUNDLE_PREFIX):
                raise ValueError("datagram is not a bundle")
            timestamp, element_datagrams = _osc_native.decode_bundle(datagram)
            contents: list[OscBundle | OscMessage] = []
            for elem in element_datagrams:
                if elem.startswith(BUNDLE_PREFIX):
                    contents.append(cls.from_datagram(elem))
                else:
                    contents.append(OscMessage.from_datagram(elem))
            return cls(timestamp=timestamp, contents=tuple(contents))
        # Fallback: pure Python
        if not datagram.startswith(BUNDLE_PREFIX):
            raise ValueError("datagram is not a bundle")
        remainder = datagram[8:]
        timestamp, remainder = cls._decode_date(remainder)
        contents_list: list[OscBundle | OscMessage] = []
        while len(remainder):
            length, remainder = struct.unpack(">i", remainder[:4])[0], remainder[4:]
            item: OscBundle | OscMessage
            if remainder.startswith(BUNDLE_PREFIX):
                item = cls.from_datagram(remainder[:length])
            else:
                item = OscMessage.from_datagram(remainder[:length])
            contents_list.append(item)
            remainder = remainder[length:]
        osc_bundle = cls(timestamp=timestamp, contents=tuple(contents_list))
        return osc_bundle

    def to_datagram(self, realtime: bool = True) -> bytes:
        datagram: bytes = BUNDLE_PREFIX
        datagram += self._encode_date(self.timestamp, realtime=realtime)
        for content in self.contents:
            content_datagram = content.to_datagram()
            datagram += struct.pack(">i", len(content_datagram))
            datagram += content_datagram
        return datagram

    def to_list(self) -> list[Any]:
        result: list[Any] = [self.timestamp]
        result.append([x.to_list() for x in self.contents])
        return result

    def to_osc(self) -> "OscBundle":
        return self


def find_free_port() -> int:
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return int(s.getsockname()[1])



"""Enum types for nanosynth."""

import enum
from collections.abc import Sequence as SequenceABC
from typing import SupportsFloat, SupportsInt, cast


class CalculationRate(enum.IntEnum):
    SCALAR = 0
    CONTROL = 1
    AUDIO = 2
    DEMAND = 3

    @classmethod
    def from_expr(cls, expr: object) -> "CalculationRate":
        if expr is None:
            return cls.SCALAR
        if isinstance(expr, cls):
            return expr
        if hasattr(expr, "calculation_rate"):
            return cast("CalculationRate", expr.calculation_rate)
        if isinstance(expr, ParameterRate):
            return {
                ParameterRate.AUDIO: cls.AUDIO,
                ParameterRate.CONTROL: cls.CONTROL,
                ParameterRate.SCALAR: cls.SCALAR,
                ParameterRate.TRIGGER: cls.CONTROL,
            }[expr]
        if isinstance(expr, (int, float, SupportsFloat)):
            return cls.SCALAR
        if isinstance(expr, str):
            return cls[expr.upper()]
        if isinstance(expr, SequenceABC):
            return max(cls.from_expr(item) for item in expr)
        return cls(int(cast(SupportsInt, expr)))

    @property
    def token(self) -> str:
        return {0: "ir", 1: "kr", 2: "ar", 3: "dr"}.get(self.value, "new")


class ParameterRate(enum.IntEnum):
    SCALAR = 0
    TRIGGER = 1
    AUDIO = 2
    CONTROL = 3

    @classmethod
    def from_expr(cls, expr: object) -> "ParameterRate":
        if expr is None:
            return cls.CONTROL
        if isinstance(expr, cls):
            return expr
        if isinstance(expr, str):
            token_map = {"ar": cls.AUDIO, "kr": cls.CONTROL, "ir": cls.SCALAR}
            lower = expr.lower()
            if lower in token_map:
                return token_map[lower]
            return cls[expr.upper()]
        return cls(int(cast(SupportsInt, expr)))


class BinaryOperator(enum.IntEnum):
    ADDITION = 0
    SUBTRACTION = 1
    MULTIPLICATION = 2
    FLOAT_DIVISION = 4
    MODULO = 5
    LESS_THAN = 8
    GREATER_THAN = 9

    @classmethod
    def from_expr(cls, expr: object) -> "BinaryOperator":
        if isinstance(expr, cls):
            return expr
        return cls(int(cast(SupportsInt, expr)))


class UnaryOperator(enum.IntEnum):
    NEGATIVE = 0
    ABSOLUTE_VALUE = 5

    @classmethod
    def from_expr(cls, expr: object) -> "UnaryOperator":
        if isinstance(expr, cls):
            return expr
        return cls(int(cast(SupportsInt, expr)))


class DoneAction(enum.IntEnum):
    NOTHING = 0
    PAUSE_SYNTH = 1
    FREE_SYNTH = 2
    FREE_SYNTH_AND_PRECEDING_NODE = 3
    FREE_SYNTH_AND_FOLLOWING_NODE = 4
    FREE_SYNTH_AND_ALL_SIBLING_NODES = 13
    FREE_SYNTH_AND_ENCLOSING_GROUP = 14


class EnvelopeShape(enum.IntEnum):
    STEP = 0
    LINEAR = 1
    EXPONENTIAL = 2
    SINE = 3
    WELCH = 4
    CUSTOM = 5
    SQUARED = 6
    CUBED = 7
    HOLD = 8

    @classmethod
    def from_expr(cls, expr: object) -> "EnvelopeShape":
        if expr is None:
            return cls.LINEAR
        if isinstance(expr, cls):
            return expr
        if isinstance(expr, str):
            return cls[expr.upper()]
        return cls(int(cast(SupportsInt, expr)))
